import sys
import json
import os.path
from typing import Union, IO
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
import lxml.etree
from codemeta.common import AttribDict, add_triple, CODEMETA, SOFTWARETYPES, add_authors, SDO, COMMON_SOURCEREPOS, SOFTWARETYPES, license_to_spdx, generate_uri
from codemeta.crosswalk import readcrosswalk, CWKey

POM_NAMESPACE = "http://maven.apache.org/POM/4.0.0"

def parse_node(node):
    for subnode in node:
        if isinstance(subnode.tag, str) and subnode.tag.startswith("{" + POM_NAMESPACE + "}"):
            key = subnode.tag[len(POM_NAMESPACE) + 2:]
            yield key, subnode

def parse_author(g, res, node, property=SDO.author):
    author_name = ""
    author_mail = ""
    author_url = ""
    org = ""
    for key3, node3 in parse_node(node):
        if key3 == "name":
            author_name = node3.text
        elif key3 == "email":
            author_mail = node3.text
        elif key3 == "url":
            author_url = node3.text
        elif key3 == "organisation":
            org = node3.text
    return add_authors(g, res, author_name, property=property, single_author=True, mail=author_mail, url=author_url,org=org)


def parse_java(g: Graph, res: Union[URIRef, BNode], file: IO , crosswalk, args: AttribDict):
    data = lxml.etree.parse(file)

    root = data.getroot()
    if root.tag != "{" + POM_NAMESPACE + "}project":
        raise Exception(f"Expected root tag 'project' in {POM_NAMESPACE} namespace, got {root.tag} instead")

    group_id = None
    artifact_id = None

    g.add((res, SDO.runtimePlatform, Literal("Java")))
    g.add((res, SDO.programmingLanguage, Literal("Java")))

    for key, node in parse_node(root):
        if key == "licenses":
            for key2, node2 in parse_node(node):
                if key2 == "license":
                    license = ""
                    for key3, node3 in parse_node(node2):
                        if key3 == "name":
                            license = license_to_spdx(node3.text)
                        if isinstance(license, str) and license.find("spdx") == -1 and key3 == "url":
                            license = node3.text
                    if license:
                        add_triple(g,res, "license", license, args)
        elif key == "issueManagement":
            for key2, node2 in parse_node(node):
                if key2 == "url" and '$' not in node2.text: #only if there are no variables in the url!
                    add_triple(g,res, "issueTracker", node2.text, args)
        elif key == "ciManagement":
            for key2, node2 in parse_node(node):
                if key2 == "url" and '$' not in node2.text: #only if there are no variables in the url!
                    add_triple(g,res, "contIntegration", node2.text, args)
        elif key == "scm":
            for key2, node2 in parse_node(node):
                if key2 == "url" and '$' not in node2.text: #only if there are no variables in the url!
                    add_triple(g,res, "codeRepository", node2.text, args)
        elif key == "repositories":
            for key2, node2 in parse_node(node):
                if key2 == "repository":
                    for key3, node3 in parse_node(node2):
                        if key3 == "url" and '$' not in node3.text: #only if there are no variables in the url!
                            add_triple(g,res, "repository", node3.text, args)
        elif key == "properties":
            for key2, node2 in parse_node(node):
                if key2 == "java.version":
                    g.add((res, SDO.runtimePlatform, Literal("Java " + node2.text)))
        elif key == 'groupId':
            group_id = node.text
        elif key == 'artifactId':
            artifact_id = node.text
        elif key == 'dependencies':
            for key2, node2 in parse_node(node):
                if key2 == "dependency":
                    dep_group_id = dep_art_id = dep_version = ""
                    for key3, node3 in parse_node(node2):
                        if key3 == "groupId":
                            dep_group_id = node3.text
                        elif key3 == "artifactId":
                            dep_art_id = node3.text
                        elif key3 == "version" and node3.text and not node3.text.startswith('$'):
                            dep_version = node3.text

                    if dep_group_id and dep_art_id:
                        depres = URIRef(generate_uri(dep_group_id +"." + dep_art_id + "." + dep_version.replace(" ",""), baseuri=args.baseuri,prefix="dependency"))
                        g.add((depres, SDO.identifier, Literal(dep_group_id + "." + dep_art_id)))
                        g.add((depres, SDO.name, Literal(dep_art_id)))
                        if dep_version:
                            g.add((depres, SDO.version, Literal(dep_version)))
                        g.add((depres, RDF.type, SDO.SoftwareApplication))
                        g.add((res, CODEMETA.softwareRequirements, depres))
        elif key == 'developers':
            for key2, node2 in parse_node(node):
                if key2 == "developer":
                    parse_author(g, res, node2)
        elif key == 'contributors':
            for key2, node2 in parse_node(node):
                if key2 == "contributor":
                    parse_author(g, res, node2, property=SDO.contributor)
        elif key == 'mailingLists':
            for key2, node2 in parse_node(node):
                if key2 == "mailingList":
                    for key3, node3 in parse_node(node2):
                        if key3 == "post":
                            add_triple(g, res, "email", node3.text, args)
        elif key == 'organization':
            org_name = None
            org_url = None
            for key2, node2 in parse_node(node):
                if key2 == "name":
                    org_name = node2.text
                elif key2 == "url":
                    org_url = node2.text
            if org_name:
                orgres = URIRef(generate_uri(org_name, baseuri=args.baseuri, prefix="org"))
                g.add((orgres, SDO.name, Literal(org_name)))
                if org_url:
                    g.add((orgres, SDO.url, Literal(org_url)))
                g.add((res, SDO.producer, orgres))
        elif key.lower() in crosswalk[CWKey.MAVEN]:
            value = node.text
            if group_id and value.find("${project.groupId}") != -1:
                value = value.replace("${project.groupId}", group_id)
            if artifact_id and value.find("${project.artifactId}") != -1:
                value = value.replace("${project.artifactId}", artifact_id)
            key = crosswalk[CWKey.MAVEN][key.lower()]
            if key != 'identifier':
                add_triple(g, res, key, value, args)

    if group_id and artifact_id:
        add_triple(g, res, "identifier", group_id + "." + artifact_id, args)



import sys
import os
import json
import requests
import random
import re
from collections import Counter, defaultdict
from tempfile import gettempdir
from rdflib import Graph, Namespace, URIRef, BNode, Literal
from rdflib.namespace import RDF, RDFS, SKOS
from rdflib.compare import graph_diff
from typing import Union, IO, Sequence, Optional,Generator
from collections import OrderedDict
from nameparser import HumanName


PROGLANG_PYTHON = {
    "@type": "ComputerLanguage",
    "name": "Python",
    "version": str(sys.version_info.major) + "." + str(sys.version_info.minor) + "." + str(sys.version_info.micro),
    "url": "https://www.python.org",
}

SDO = Namespace("http://schema.org/")

CODEMETA = Namespace("https://codemeta.github.io/terms/")
#Custom extensions not in codemeta/schema.org (yet), they are initially proposed in https://github.com/codemeta/codemeta/issues/271
SOFTWARETYPES = Namespace("https://w3id.org/software-types#") #See https://github.com/SoftwareUnderstanding/software_types
SOFTWAREIODATA = Namespace("https://w3id.org/software-iodata#") #See https://github.com/SoftwareUnderstanding/software-iodata
TRL = Namespace("https://w3id.org/research-technology-readiness-levels#")

REPOSTATUS = Namespace("https://www.repostatus.org/#")

SPDX = Namespace("http://spdx.org/licenses/")

ORCID = Namespace("http://orcid.org/")

CODEMETAPY = Namespace("https://github.com/proycon/codemetapy/") #An extra internal namespace, usually not used for any serialisations

SCHEMA_SOURCE = "https://schema.org" #even though URL is https:// the RDF IRIs are all http://!
CODEMETA_SOURCE = "https://doi.org/10.5063/schema/codemeta-2.0"
STYPE_SOURCE = "https://w3id.org/software-types"
IODATA_SOURCE = "https://w3id.org/software-iodata"

REPOSTATUS_SOURCE = "https://raw.githubusercontent.com/jantman/repostatus.org/master/badges/latest/ontology.jsonld"

TMPDIR  = os.environ.get("TMPDIR",gettempdir())

SCHEMA_LOCAL_SOURCE = "file://" + os.path.join(TMPDIR, "schemaorgcontext.jsonld")
CODEMETA_LOCAL_SOURCE = "file://" + os.path.join(TMPDIR, "codemeta.jsonld")
STYPE_LOCAL_SOURCE = "file://" + os.path.join(TMPDIR, "stype.jsonld")
IODATA_LOCAL_SOURCE = "file://" + os.path.join(TMPDIR, "iodata.jsonld")
REPOSTATUS_LOCAL_SOURCE = "file://" + os.path.join(TMPDIR, "repostatus.jsonld")

COMMON_SOURCEREPOS = ["https://github.com/","http://github.com","https://gitlab.com/","http://gitlab.com/","https://codeberg.org/","http://codeberg.org", "https://git.sr.ht/", "https://bitbucket.org/", "https://bitbucket.com/"]


ENTRYPOINT_CONTEXT = { #these are all custom extensions not in codemeta (yet), they were proposed in https://github.com/codemeta/codemeta/issues/183 but are obsolete in favour of the newer software types
    "entryPoints": { "@reverse": "schema:actionApplication" },
    "interfaceType": { "@id": "codemeta:interfaceType" }, #Type of the entrypoint's interface (e.g CLI, GUI, WUI, TUI, REST, SOAP, XMLRPC, LIB)
    "specification": { "@id": "codemeta:specification" , "@type":"@id"}, #A technical specification of the interface
    "mediatorApplication": {"@id": "codemeta:mediatorApplication", "@type":"@id" } #auxiliary software that provided/enabled this entrypoint
}


DEVIANT_CONTEXT = { #Extra *internal* context that enforces some implementation-specific differences and solves conflicts, this context is ALWAYS assumed on parsing, and ALWAYS passed to the serialiser (but NEVER propagated to final output, it is stripped again on final context rewrite) 
    #"softwareRequirements":{ "@id": "schema:softwareRequirements","@type":"@id"}, #schema does not add @type=@id, codemeta does....   
    #"referencePublication":{ "@id": "codemeta:referencePublication","@type":"@id"}, #schema does not add @type=@id, codemeta does....
    "author": {"@id": "schema:author", "@container": "@list" },  #we treat authors as an ordered list (rdf:list), even though schema.org does not (this is also recommended by science-on-schema.org)
    "contributor": {"@id": "schema:contributor", "@container": "@list" },  #we treat contributors as an ordered list (rdf:list), even though schema.org does not (this is also recommended by science-on-schema.org)
}


REPOSTATUS_MAP = { #maps Python development status to repostatus.org vocabulary (the mapping is debatable)
    "1 - planning": "concept",
    "2 - pre-alpha": "concept",
    "3 - alpha": "wip",
    "4 - beta": "wip", #note, if --released is set this maps to "active" instead
    "5 - production/stable": "active", #reasonable guess, but being production/stable doesn't guarantee repo is active (might be inactive)
    "6 - mature": "active", #reasonable guess, but being mature doesn't guarantee repo is active (might be inactive)
    "7 - inactive": "unsupported",
}

TRL_MAP = { #maps Python development status to technology readiness levels (the mapping is debatable)
    "1 - planning": TRL.Stage1Planning,
    "2 - pre-alpha": TRL.Stage1Planning, #we can't really distinguish between the TRL levels in stage 1
    "3 - alpha": TRL.Stage2ProofOfConcept,
    "4 - beta": TRL.Stage3Experimental,
    "5 - production/stable": TRL.Level8Complete,
    "6 - mature": TRL.Level9Proven
    #7 - inactive does not map to a TRL
}

LICENSE_MAP = [ #maps some common licenses to SPDX URIs, mapped with a substring match on first come first serve basis
    ("GNU General Public License v3.0 or later", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GNU General Public License v3 or later", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GNU General Public License v3", "http://spdx.org/licenses/GPL-3.0-only"),
    ("GNU General Public License 3.0 or later", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GNU General Public License 3 or later", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GNU General Public License 3", "http://spdx.org/licenses/GPL-3.0-only"),
    ("GNU General Public License v2.0 or later", "http://spdx.org/licenses/GPL-2.0-or-later"),
    ("GNU General Public License v2 or later", "http://spdx.org/licenses/GPL-2.0-or-later"),
    ("GNU General Public License v2", "http://spdx.org/licenses/GPL-2.0-only"),
    ("GNU General Public License 2.0 or later", "http://spdx.org/licenses/GPL-2.0-or-later"),
    ("GNU General Public License 2 or later", "http://spdx.org/licenses/GPL-2.0-or-later"),
    ("GNU General Public License 2", "http://spdx.org/licenses/GPL-2.0-only"),
    ("GNU Affero General Public License v3.0 or later", "http://spdx.org/licenses/AGPL-3.0-or-later"),
    ("GNU Affero General Public License v3 or later", "http://spdx.org/licenses/AGPL-3.0-or-later"),
    ("GNU Affero General Public License v3", "http://spdx.org/licenses/AGPL-3.0-only"),
    ("GNU Affero General Public License", "http://spdx.org/licenses/AGPL-3.0-only"),
    ("GNU Lesser General Public License v3.0 or later", "http://spdx.org/licenses/LGPL-3.0-or-later"),
    ("GNU Lesser General Public License v3", "http://spdx.org/licenses/LGPL-3.0-only"),
    ("GNU Lesser General Public License v2.1 or later", "http://spdx.org/licenses/LGPL-2.1-or-later"),
    ("GNU Lesser General Public License v2.1", "http://spdx.org/licenses/LGPL-2.1-only"),
    ("GNU Lesser General Public License 2.1", "http://spdx.org/licenses/LGPL-2.1-only"),
    ("GNU Lesser General Public License v2 or later", "http://spdx.org/licenses/LGPL-2.0-or-later"),
    ("GNU Lesser General Public License v2", "http://spdx.org/licenses/LGPL-2.0-only"),
    ("Mozilla Public License 1.1", "http://spdx.org/licenses/MPL-1.1"),
    ("Mozilla Public License 1", "http://spdx.org/licenses/MPL-1.0"),
    ("Mozilla Public License", "http://spdx.org/licenses/MPL-2.0"),
    ("https://www.mozilla.org/en-US/MPL/2.0/","http://spdx.org/licenses/MPL-2.0"),
    ("https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12", "http://spdx.org/licenses/EUPL-1.2"),
    ("European Union Public License 1.1", "http://spdx.org/licenses/EUPL-1.1"),
    ("European Union Public License", "http://spdx.org/licenses/EUPL-1.2"),
    ("Eclipse Public License 1", "http://spdx.org/licenses/EPL-1.0"),
    ("Eclipse Public License", "http://spdx.org/licenses/EPL-2.0"),
    ("Common Public Attribution License", "http://spdx.org/licenses/CPAL-1.0"),
    ("Apache License 2", "http://spdx.org/licenses/Apache-2.0"),
    ("Apache License, Version 2.0", "http://spdx.org/licenses/Apache-2.0"),
    ("Apache License", "http://spdx.org/licenses/Apache-1.1"),
    ("https://www.apache.org/licenses/LICENSE-2.0.txt", "http://spdx.org/licenses/Apache-2.0"),
    ("http://www.apache.org/licenses/LICENSE-2.0.txt", "http://spdx.org/licenses/Apache-2.0"),
    ("https://www.apache.org/licenses/LICENSE-2.0.html", "http://spdx.org/licenses/Apache-2.0"),
    ("http://www.apache.org/licenses/LICENSE-2.0.html", "http://spdx.org/licenses/Apache-2.0"),
    ("Apache-2.0", "http://spdx.org/licenses/Apache-2.0"),
    ("Apache-1.1", "http://spdx.org/licenses/Apache-1.1"),
    ("Apache", "http://spdx.org/licenses/Apache-2.0"), #ambiguous, assume apache 2.0
    ("https://www.gnu.org/licenses/agpl-3.0.html","http://spdx.org/licenses/AGPL-3.0-only"),
    ("http://www.gnu.org/licenses/agpl-3.0.html","http://spdx.org/licenses/AGPL-3.0-only"),
    ("AGPL-3.0-or-later", "http://spdx.org/licenses/AGPL-3.0-or-later"),
    ("AGPL-3.0-only", "http://spdx.org/licenses/AGPL-3.0-only"),
    ("AGPL-3.0", "http://spdx.org/licenses/AGPL-3.0-only"),
    ("https://www.gnu.org/licenses/gpl-3.0.html","http://spdx.org/licenses/GPL-3.0-only"),
    ("http://www.gnu.org/licenses/gpl-3.0.html","http://spdx.org/licenses/GPL-3.0-only"),
    ("GPL-3.0-or-later", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GPL-3.0-only", "http://spdx.org/licenses/GPL-3.0-only"),
    ("GPL-3.0+", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GPL-3.0", "http://spdx.org/licenses/GPL-3.0-only"),
    ("GPLv3+", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GPLv3", "http://spdx.org/licenses/GPL-3.0-only"),
    ("GPL3+", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GPL3", "http://spdx.org/licenses/GPL-3.0-only"),
    ("https://www.gnu.org/licenses/gpl-2.0.html","http://spdx.org/licenses/GPL-2.0-only"),
    ("http://www.gnu.org/licenses/gpl-2.0.html","http://spdx.org/licenses/GPL-2.0-only"),
    ("GPL-2.0-or-later", "http://spdx.org/licenses/GPL-2.0-or-later"),
    ("GPL-2.0-only", "http://spdx.org/licenses/GPL-2.0-only"),
    ("GPL-2.0", "http://spdx.org/licenses/GPL-2.0-only"),
    ("GPLv2+", "http://spdx.org/licenses/GPL-2.0-or-later"),
    ("GPLv2", "http://spdx.org/licenses/GPL-2.0-only"),
    ("GPL3+", "http://spdx.org/licenses/GPL-3.0-or-later"),
    ("GPL3", "http://spdx.org/licenses/GPL-3.0-only"),
    ("GPL2+", "http://spdx.org/licenses/GPL-2.0-or-later"),
    ("GPL2", "http://spdx.org/licenses/GPL-2.0-only"),
    ("GPL", "http://spdx.org/licenses/GPL-2.0-or-later"), #rather ambiguous, assuming all that could apply
    ("BSD-3-Clause", "http://spdx.org/licenses/BSD-3-Clause"),
    ("BSD-2-Clause", "http://spdx.org/licenses/BSD-2-Clause"),
    ("Simplified BSD", "http://spdx.org/licenses/BSD-2-Clause"),
    ("FreeBSD", "http://spdx.org/licenses/BSD-2-Clause"),
    ("BSD License", "http://spdx.org/licenses/BSD-3-Clause"), #may be ambiguous, better be as restrictive
    ("BSD", "http://spdx.org/licenses/BSD-3-Clause"), #may be ambiguous, better be as restrictive
    ("MIT No Attribution", "http://spdx.org/licenses/MIT-0"),
    ("MIT", "http://spdx.org/licenses/MIT"),
    ("Creative Commons Attribution Share Alike 4.0 International", "http://spdx.org/licenses/CC-BY-SA-4.0"), #not designed for software, not OSI-approved
    ("https://creativecommons.org/licenses/by-sa/4.0","http://spdx.org/licenses/CC-BY-SA-4.0"),
    ("CC-BY-SA-4.0", "http://spdx.org/licenses/CC-BY-SA-4.0"), #not designed for software, not OSI-approved
]

#Maps invalid characters in URIs to something sensible (especially characters from dependency version qualifiers)
IDENTIFIER_MAP = [
   ('>=', '-ge-'),
   ('~=', '-ge-'),
   ('>=', '-le-'),
   ('==', '-eq-'),
   ('!=', '-ne-'),
   ('^', '-ge-'), #used in npm version
   ('>', '-gt-'),
   ('<', '-lt-'),
   ('=', '-eq-'),
   (' ', '-'),
   ('&', '-',),
   ('/', '-',),
   ('+', '-',),
   (':', '-',),
   (';', '-'),
   (',',''),
   ('----', '-'),
   ('---', '-'),
   ('--', '-'),
]

#keywords that may be indicative of a certain interface type
INTERFACE_CLUES = [ #order matters, only one is picked
   ("web application", SDO.WebApplication),
   ("webapp", SDO.WebApplication),
   ("web-based", SDO.WebApplication),
   ("website", SDO.WebSite),
   ("webpage", SDO.WebPage),
   ("web service", SDO.WebAPI),
   ("webservice", SDO.WebAPI),
   ("restful", SDO.WebAPI),
   ("rest service", SDO.WebAPI),
   ("web api", SDO.WebAPI),
   ("library", SOFTWARETYPES.SoftwareLibrary),
   ("module", SOFTWARETYPES.SoftwareLibrary),
   ("command-line", SOFTWARETYPES.CommandLineApplication),
   ("command line", SOFTWARETYPES.CommandLineApplication),
   ("commandline", SOFTWARETYPES.CommandLineApplication),
   ("desktop application", SOFTWARETYPES.DesktopApplication),
   ("windows application", SOFTWARETYPES.DesktopApplication),
   ("windows software", SOFTWARETYPES.DesktopApplication),
   ("mac application", SOFTWARETYPES.DesktopApplication),
   ("graphical user-interface", SOFTWARETYPES.DesktopApplication),
   ("graphical user interface", SOFTWARETYPES.DesktopApplication),
   ("gnome", SOFTWARETYPES.DesktopApplication),
   ("gtk+", SOFTWARETYPES.DesktopApplication),
   (" qt ", SOFTWARETYPES.DesktopApplication),
   (" gui", SOFTWARETYPES.DesktopApplication),
   ("desktop gui", SOFTWARETYPES.DesktopApplication),
   ("android app", SOFTWARETYPES.MobileApplication),
   ("ios app", SOFTWARETYPES.MobileApplication),
   ("mobile app", SOFTWARETYPES.MobileApplication),
   ("in a terminal", SOFTWARETYPES.CommandLineApplication),
   ("in the terminal", SOFTWARETYPES.CommandLineApplication),
   ("from the terminal", SOFTWARETYPES.CommandLineApplication),
   ("from a terminal", SOFTWARETYPES.CommandLineApplication),
   (" api ", SOFTWARETYPES.SoftwareLibrary)
]

INTERFACE_CLUES_DEPS = {
    "django": SOFTWARETYPES.WebApplication,
    "flask": SOFTWARETYPES.WebApplication,
    "clam": SOFTWARETYPES.WebApplication,
    "react": SOFTWARETYPES.WebApplication,
    "vue": SOFTWARETYPES.WebApplication,
    "jquery": SOFTWARETYPES.WebApplication,
    "gatsby": SOFTWARETYPES.WebApplication,
    "angular": SOFTWARETYPES.WebApplication,
    "laravel": SOFTWARETYPES.WebApplication,
    "drupal": SOFTWARETYPES.WebApplication,
    "wordpress": SOFTWARETYPES.WebApplication,
    "joomla": SOFTWARETYPES.WebApplication,
    "spring": SOFTWARETYPES.WebApplication,
    "web2py": SOFTWARETYPES.WebApplication,
    "fastapi": SDO.WebAPI, #the distinction webservice/webapplication on the basis of dependencies is fairly ambiguous
    "bottle": SDO.WebAPI,
    "hug": SDO.WebAPI,
    "falcon": SDO.WebAPI,
    "tornado": SDO.WebAPI,
    "cherrypy": SDO.WebAPI,
    "ncurses": SOFTWARETYPES.TerminalApplication,
    "textual": SOFTWARETYPES.TerminalApplication,
    "click": SOFTWARETYPES.CommandLineApplication
}



#properties that may only occur once, last one counts
SINGULAR_PROPERTIES = ( SDO.name, SDO.version, SDO.description, SDO.dateCreated, SDO.dateModified, SDO.position )

#these will be treated as ordered lists
ORDEREDLIST_PROPERTIES = ( SDO.author, SDO.contributor )

#properties that should prefer URIRef rather than Literal **if and only if** the value is a URI, even though the @context might not make this explicit (e.g. interpret as if @type=@id)
PREFER_URIREF_PROPERTIES = (SDO.license,  CODEMETA.developmentStatus )



class AttribDict(dict):
    """Simple dictionary that is addressable via attributes"""
    def __init__(self, d: dict):
        self.__dict__ = d

    def __getattr__(self, key):
        if key in self:
            return self[key]
        return None

def init_context(args: AttribDict):
    """Initialized the context, ensures all context JSONLDs are downloaded and local filesystem references are used instead"""

    sources = [ (CODEMETA_LOCAL_SOURCE, CODEMETA_SOURCE), (SCHEMA_LOCAL_SOURCE, SCHEMA_SOURCE), (STYPE_LOCAL_SOURCE, STYPE_SOURCE), (IODATA_LOCAL_SOURCE, IODATA_SOURCE), (REPOSTATUS_LOCAL_SOURCE, REPOSTATUS_SOURCE) ]
    
    if args.addcontext:
        for remote_url in args.addcontext:
            if not remote_url.startswith("http"):
                raise Exception(f"Explicitly added context (--addcontext) must be a remote URL, got {remote_url} instead")
            local = "file://" + os.path.join(TMPDIR, os.path.basename(remote_url))
            sources.append( (local, remote_url))

    for local, remote in sources:
        localfile = local.replace("file://","")
        if remote in ("http://schema.org", "https://schema.org","http://schema.org/", "https://schema.org/"):
            #schema.org does not do content negotation properly, instead it provides a link via a HEAD request, we don't support this but fake this step manually:
            remote = "https://schema.org/docs/jsonldcontext.json"
        if not os.path.exists(localfile) or args.no_cache:
            print(f"Downloading context from {remote}", file=sys.stderr)
            if remote.find("doi.org") != -1:
                #if we use application/ld+json on doi.org URL we get metadata of the DOI resource itself rather than the jsonld it references (relevant for codemeta)
                accept = "application/json;q=0.9,text/plain;q=0.5"
            else:
                accept = "application/ld+json;q=1.0;application/json;q=0.9,text/plain;q=0.5"
            r = requests.get(remote, headers={ "Accept": accept})
            r.raise_for_status()
            with open(localfile, 'wb') as f:
                f.write(r.content)

    return sources

def bind_graph(g: Graph):
    g.bind('schema', SDO)
    g.bind('codemeta', CODEMETA)
    g.bind('stype', SOFTWARETYPES)
    g.bind('iodata', SOFTWAREIODATA)

def init_graph(args: AttribDict):
    """Initializes the RDF graph, the context and the prefixes"""

    context_sources = init_context(args)

    g = Graph()


    #The context graph loads some additional linked data we may need for interpretation (it is not related to @context!),
    #This data is not propagated to the output graph (g) unless --includecontext is set

    contextgraph = Graph()
    for x in (g, contextgraph):
        bind_graph(x)

    contextgraph.bind('rdfs', RDFS)
    contextgraph.bind('repostatus', REPOSTATUS)
    contextgraph.bind('spdx', SPDX)
    contextgraph.bind('skos', SKOS)
    contextgraph.bind('orcid', ORCID)
    contextgraph.bind('trl', TRL)

    #add license names from our license map (faster/easier than ingesting the json-ld from https://github.com/spdx/license-list-data/)
    for (label, identifier) in LICENSE_MAP:
        license = URIRef(identifier)
        if (license, SDO.name, None) not in contextgraph and not label.startswith("http"):
            contextgraph.add((license, SDO.name, Literal(label)))

    #Add labels for software types that are not in the software types extension but in schema.org itself (without needing to parse everythign in schema.org)
    contextgraph.add((SDO.WebApplication, RDFS.label, Literal("Web Application")))
    contextgraph.add((SDO.WebApplication, RDFS.comment, Literal("A software application served as a service over the web with an interface for human end-users")))
    contextgraph.add((SDO.WebAPI, RDFS.label, Literal("Web API")))
    contextgraph.add((SDO.WebApplication, RDFS.comment, Literal("A software application served as a service over the web with an interface for human end-users")))
    contextgraph.add((SDO.MobileApplication, RDFS.label, Literal("Mobile App")))
    contextgraph.add((SDO.MobileApplication, RDFS.comment, Literal("A software application for mobile devices")))
    contextgraph.add((SDO.WebSite, RDFS.label, Literal("Website")))
    contextgraph.add((SDO.WebSite, RDFS.comment, Literal("A set of related web pages")))
    contextgraph.add((SDO.WebPage, RDFS.label, Literal("Webpage")))
    contextgraph.add((SDO.WebPage, RDFS.comment, Literal("A very particular page on the web")))

    for local, _ in context_sources:
        with open(local.replace("file://",""),'rb') as f:
            contextgraph.parse(data=json.load(f), format="json-ld")

    if args.addcontextgraph:
        #these are only added to the context graph, but NOT the json-ld context, here you can also add turtle files
        for url in args.addcontextgraph:
            if url.startswith("http"):
                localfile = os.path.join(TMPDIR, os.path.basename(url))
            else:
                localfile = url 
            if not os.path.exists(localfile) and url.startswith("http"):
                print(f"Downloading data for contextgraph from {url}", file=sys.stderr)
                accept = "application/ld+json;q=1.0;application/json;q=0.9;text/turtle;q=0.8,text/plain;q=0.5"
                r = requests.get(url, headers={ "Accept": accept})
                r.raise_for_status()
                with open(localfile, 'wb') as f:
                    f.write(r.content)
            print(f"Adding to contextgraph: {localfile}", file=sys.stderr)
            with open(localfile, 'r') as f:
                contextgraph.parse(f)

    return g, contextgraph


def license_to_spdx(value: Union[str,list,tuple]) -> Union[str,list]:
    """Attempts to converts a license name or acronym to a full SPDX URI (https://spdx.org/licenses/)"""
    if isinstance(value, (list,tuple)):
        return [ license_to_spdx(x) for x in value ]
    if value in ("http://spdx.org/licenses/GPL-3.0", "https://spdx.org/licenses/GPL-3.0"):
        #short form is too inprecise and deprecated, resolve to most restrictive form:
        return "http://spdx.org/licenses/GPL-3.0-only"
    if value in ("http://spdx.org/licenses/GPL-2.0"):
        #short form is too inprecise and deprecated, resolve to most restrictive form:
        return "http://spdx.org/licenses/GPL-2.0-only"
    if value.startswith("http://spdx.org"):
        #we're already good, nothing to do
        return value
    if value.startswith("https://spdx.org"):
        #we consistently opt for http:// in this implementation
        return value.replace("https://","http://")
    if value.startswith("https://opensource.org/licenses/"): #OSI
        #extract license identifier from base (this is usually SPDX and will resolve in the next step)
        value = value[len("https://opensource.org/licenses/"):]
        if value in ("GPL-3.0","GPL-2.0","AGPL-3.0","LGPL-2.1"):
            #short form is too inprecise and deprecated, resolve to most restrictive form:
            value += "-only"
    for substr, license_uri in LICENSE_MAP:
        if value.find(substr) != -1:
            return license_uri
    return value

def detect_list(value: Union[list,tuple,set,str]) -> Union[list,str]:
    """Tries to see if the value is a list (string of comma separated items) and then returns a list"""
    if isinstance(value, (list,tuple)):
        return value
    if isinstance(value, str) and value.count(',') >= 1:
        return [ x.strip() for x in value.split(",") ]
    return value


def resolve(data: dict, idmap=None) -> dict:
    """Resolve nodes that refer to an ID mentioned earlier"""
    if idmap is None: idmap = {}
    for k,v in data.items():
        if isinstance(v, (dict, OrderedDict)):
            if '@id' in v:
                if len(v) > 1:
                    #this is not a reference, register in idmap (possibly overwriting earlier definition!)
                    idmap[v['@id']] = v
                elif len(v) == 1:
                    #this is a reference
                    if v['@id'] in idmap:
                        data[k] = idmap[v['@id']]
                    else:
                        print("NOTICE: Unable to resolve @id " + v['@id'] ,file=sys.stderr)
            resolve(v, idmap)
        elif isinstance(v, (tuple, list)):
            data[k] = [ resolve(x,idmap) if isinstance(x, (dict,OrderedDict)) else x for x in v ]
    return data

def getregistry(identifier, registry):
    for tool in registry['@graph']:
        if tool['identifier'].lower() == identifier.lower():
            return tool
    raise KeyError(identifier)


def delete_repostatus(g: Graph, res: Union[URIRef, BNode]):
    """Delete any existing developmentStatus repostatus triples before adding a new one (there may be only one)"""
    for _,_,o in g.triples((res, CODEMETA.developmentStatus, None)):
        if str(o).find("repostatus") != -1:
            g.remove((res,CODEMETA.developmentStatus, o))

def add_triple(g: Graph, res: Union[URIRef, BNode],key, value, args: AttribDict, replace=False) -> bool:
    """Maps a key/value pair to an actual triple"""

    if replace or key in ( x.split("/")[-1] for x in SINGULAR_PROPERTIES  ):
        f_add = g.set
    else:
        f_add = g.add
    if key == "developmentStatus":
        f_add = g.add
        if value.strip().lower() in REPOSTATUS_MAP.values():
            delete_repostatus(g, res)
            f_add((res, CODEMETA.developmentStatus, getattr(REPOSTATUS, value.strip().lower()) ))
        elif value.strip().lower() in REPOSTATUS_MAP:
            #map to repostatus vocabulary
            repostatus = REPOSTATUS_MAP[value.strip().lower()]
            if args.released and value.strip().lower().find("beta") and repostatus == "wip":
                #beta maps to active if --released is set
                repostatus = "active"
            delete_repostatus(g, res)
            f_add((res, CODEMETA.developmentStatus, getattr(REPOSTATUS, repostatus) ))
        else:
            delete_repostatus(g, res)
            f_add((res, CODEMETA.developmentStatus, Literal(value)))
        if args.trl:
            if value.strip().lower() in TRL_MAP:
                f_add((res, CODEMETA.developmentStatus, TRL_MAP[value.strip().lower()] ))
    elif key == "license": 
        if value in ( "UNKNOWN","NOASSERTION"):
            #python distutils has a tendency to assign 'UNKNOWN', we don't use this value
            #npm uses NOASSERTION?
            return True
        value = license_to_spdx(value)
        if isinstance(value, str):
            listify = lambda x: [x]
        else:
            listify = lambda x: x
        for value in listify(value):
            if value.find('spdx') != -1:
                f_add((res, SDO.license, URIRef(value)))
            else:
                f_add((res, SDO.license, Literal(value)))
    elif key == "applicationCategory":
        f_add((res, SDO.applicationCategory, Literal(value)))
    elif key == "audience":
        audience = URIRef(generate_uri(value, baseuri=args.baseuri, prefix="audience"))
        g.add((audience, RDF.type, SDO.Audience))
        g.add((audience, SDO.audienceType, Literal(value) ))
        f_add((res, SDO.audience,audience ))
    elif key == "keywords":
        value = detect_list(value)
        if isinstance(value, list):
            for item in value:
                g.add((res, SDO.keywords,Literal(item)))
        else:
            f_add((res, SDO.keywords,Literal(value)))
    elif key == "operatingSystem" and isinstance(value, str) and value.upper() in ("POSIX","UNIX"):
        #decompose into some actual operating systems that are easier to understand (Python packages often use POSIX as a group)
        f_add((res, SDO.operatingSystem, Literal("Linux")))
        f_add((res, SDO.operatingSystem, Literal("BSD")))
        f_add((res, SDO.operatingSystem, Literal("macOS")))
    elif hasattr(SDO, key):
        f_add((res, getattr(SDO, key), Literal(value)))
    elif hasattr(CODEMETA, key):
        f_add((res, getattr(CODEMETA, key), Literal(value)))
    else:
        print(f"NOTICE: Don't know how to handle key '{key}' with value '{value}'... ignoring...",file=sys.stderr)
        return False
    return True

def add_authors(g: Graph, res: Union[URIRef, BNode], value, property=SDO.author, single_author = False,  **kwargs):
    """Parse and add authors and their e-mail addresses"""
    if single_author:
        names = [value.strip()]
        mails = [ kwargs.get('mail',"") ]
        urls = [ kwargs.get('url',"") ]
        orgs = [ kwargs.get('organization',"") ]
    else:
        names = value.strip().split(",")
        mails = kwargs.get("mail","").strip().split(",")
        urls = kwargs.get('url',"").strip().split(",")
        orgs = kwargs.get('organization',"").strip().split(",")


    skip_duplicates = kwargs.get('skip_duplicates')

    authors = []
    for i, name in enumerate(names):
        if len(mails) > i:
            mail = mails[i]
        else:
            mail = None
        if len(urls) > i:
            url = urls[i]
        else:
            url = None
        if len(orgs) > i:
            org = orgs[i]
        else:
            org = None

        if not mail:
            #mails and urls may be tucked away with the name
            # npm allows strings like "Barney Rubble <b@rubble.com> (http://barnyrubble.tumblr.com/)"
            # we do the same, alternatively we allow affiliations:
            #       "Barney Rubble <b@rubble.com> (Barney's Chocolate Factory)"
            m = re.search(r'([^<]+)(?:<([^@]+@[^>]+)>)?\s*(?:(\([^\)]+\)))?',name)
            if m:
                name, mail, extra = m.groups()
                if extra:
                    if extra.startswith("http"):
                        url = extra
                    elif extra.startswith("www"):
                        url = "http://" + extra
                    else:
                        org = extra

        firstname, lastname = parse_human_name(name.strip())

        author = URIRef(generate_uri(firstname + "-" + lastname, kwargs.get('baseuri'), prefix="person"))
        if skip_duplicates:
            q = f"SELECT ?a WHERE {{ ?a a schema:Person . ?a schema:givenName \"{firstname}\" . ?a schema:familyName \"{lastname}\" . }}"
            if g.query(q, initNs={'schema': SDO }):
                #person already exists, skipping
                continue
            if mail and g.query(f"SELECT ?a WHERE {{ ?a a schema:Person . ?a schema:email \"{mail}\" . }}", initNs={ 'schema': SDO }):
                #mail already exists, skipping
                continue

        g.add((author, RDF.type, SDO.Person))
        g.add((author, SDO.givenName, Literal(firstname.strip())))
        g.add((author, SDO.familyName, Literal(lastname.strip())))
        if mail and '@' in mail:
            g.add((author, SDO.email, Literal(mail.strip())))
        if url:
            g.add((author, SDO.url, Literal(url.strip("() "))))
            #                              -------------^
            #  needed to cleanup after the regexp and to prevent other accidents
        if org:
            orgres = URIRef(generate_uri(org, kwargs.get('baseuri'), prefix="org"))
            g.add((orgres, RDF.type, SDO.Organization))
            g.add((orgres, SDO.name, Literal(org.strip())))
            g.add((author, SDO.affiliation, orgres))

        if property in ORDEREDLIST_PROPERTIES:
            add_to_ordered_list(g, res, property, author)
        else:
            g.add((res, property, author))
        authors.append(author) #return the nodes

    return authors

def add_to_ordered_list(g: Graph, subject: Union[URIRef, BNode], property: URIRef, object: Union[URIRef, BNode, Literal], identifying_properties: list = [SDO.name,SDO.email]):
    """Add an item to the end of an ordered list in RDF (rdf:first, rdf:next)"""
    collection = g.value(subject, property)

    #make an invetory of property that might identify the object, so we don't add it twice
    idpropmap = {}
    for prop in identifying_properties:
        if isinstance(object, (URIRef, BNode)):
            v = g.value(object,prop)
            if v:
                idpropmap[prop] = v

    if collection and isinstance(collection, (URIRef,BNode)) and (collection, RDF.first,None) in g:
        while True:
            if (collection, RDF.first, object) in g:
                #item already exists, nothing to add
                return False
            testobject = g.value(collection, RDF.first)
            if isinstance(testobject, (URIRef, BNode)):
                for prop, value in idpropmap.items():
                    v = g.value(testobject,prop)
                    if v and v == value:
                        #item already exists, nothing to add
                        return False
            end = g.value(collection, RDF.rest)
            if end == RDF.nil or not end:
                break
            collection = end
        if not end:
            raise Exception(f"Unable to find end of ordered list {collection}")
        g.remove((collection, RDF.rest, RDF.nil))
        newnode = BNode()
        g.add((newnode, RDF.first, object)) 
        g.add((newnode, RDF.rest, RDF.nil)) 
        g.add((collection, RDF.rest, newnode))
        return True
    elif not collection:
        collection = BNode()
        g.add((subject,property, collection))
        g.add((collection, RDF.first, object))
        g.add((collection, RDF.rest, RDF.nil))
        return True
    return False

def part_of_ordered_list(g: Graph, subject: Union[URIRef, BNode], property: URIRef, object: Union[URIRef, BNode, Literal]) -> bool:
    """Check if an item is a member of an RDF ordered list (rdf:first, rdf:rest)"""
    collection = g.value(subject, property)
    while collection and isinstance(collection, (URIRef,BNode)):
        if (collection, RDF.first, object) in g:
            return True
        collection = g.value(collection, RDF.rest)
    return False

def iter_ordered_list(g: Graph, subject: Union[URIRef, BNode], property: URIRef) -> Generator:
    """Check if an item is a member of an RDF ordered list (rdf:first, rdf:rest). Returns triples"""
    collection = g.value(subject, property)
    while collection and isinstance(collection, (URIRef,BNode)):
        object =  g.value(collection, RDF.first)
        if object:
            yield subject, property, object
        next_collection = g.value(collection, RDF.rest)
        if not object and not collection:
            #this is not an ordered list at all!
            print(f"WARNING: iter_ordered_list called but no ordered list found! Falling back to unordered iteration...", file=sys.stderr)
            for _,_,object in g.triples((subject,property,None)):
                yield subject, property, object
        collection = next_collection

def parse_human_name(name):
    humanname = HumanName(name.strip())
    lastname = " ".join((humanname.middle, humanname.last)).strip()
    return humanname.first, lastname


def get_last_component(uri):
    index = max(uri.rfind('#'), uri.rfind('/'))
    if index == -1:
        return uri
    else:
        return uri[index+1:]

def reconcile(g: Graph, res: URIRef, args: AttribDict):
    """Reconcile possible conflicts in the graph and issue warnings."""
    IDENTIFIER = g.value(res, SDO.identifier)
    if not IDENTIFIER: IDENTIFIER = str(res)
    HEAD = f"[CODEMETA VALIDATION ({IDENTIFIER})]"

    if (res, SDO.codeRepository, None) not in g:
        print(f"{HEAD} codeRepository not set",file=sys.stderr)
    if (res, SDO.author, None) not in g:
        print(f"{HEAD} author not set",file=sys.stderr)
    if (res, SDO.license, None) not in g:
        print(f"{HEAD} license not set",file=sys.stderr)

    #rewrite technology readiness level to developmentStatus
    for s,p,o in g.triples((res, TRL.technologyReadinessLevel, None)):
        if str(o).startswith(TRL):
            g.remove((s,p,o))
            g.add((s,CODEMETA.developmentStatus,o))

    if (res, SDO.license, URIRef("http://spdx.org/licenses/GPL-3.0-only")) in g and (res, SDO.license, URIRef("http://spdx.org/licenses/GPL-2.0-or-later")) in g:
        print(f"{HEAD} license conflict: found GPL-3.0-only and GPL-2.0-or-later, removing the latter",file=sys.stderr)
        g.remove((res, SDO.license, URIRef("http://spdx.org/licenses/GPL-2.0-or-later")))
    if (res, SDO.license, URIRef("http://spdx.org/licenses/GPL-3.0-or-later")) in g and (res, SDO.license, URIRef("http://spdx.org/licenses/GPL-2.0-or-later")) in g:
        print(f"{HEAD} license conflict: found GPL-3.0-or-later and GPL-2.0-or-later, removing the latter",file=sys.stderr)
        g.remove((res, SDO.license, URIRef("http://spdx.org/licenses/GPL-2.0-or-later"))) #take the more restrictive option


    #if there are multiple target products of the same type, delete any 'stub' ones that have no url, as these stub ones were created by codemetapy based on interface type and carry the same name as the SoftwareSourceCode resource. The stub entries are generated by guess_interfacetype(), called by enrich()
    name = g.value(res,SDO.name)
    if name:
        targetproducts = set()
        for _,_,targetproduct in g.triples((res, SDO.targetProduct, None)):
            if isinstance(targetproduct, (URIRef, BNode)):
                for _,_,t in g.triples((targetproduct,RDF.type,None)):
                    if t in (SDO.WebApplication, SDO.WebAPI):
                        targetproducts.add(targetproduct)
        if len(targetproducts) > 1:
            no_stubs = set(targetproduct for targetproduct in targetproducts if g.value(targetproduct, SDO.url) != None)
            stubs = set(targetproduct for targetproduct in targetproducts if not g.value(targetproduct, SDO.url) and (targetproduct, SDO.name, name) in g)
            if stubs and no_stubs:
                for stub in stubs:
                    print(f"{HEAD} removing stub targetProduct (WebApplication or WebAPI) without a URL, as we already have one (or more) with URL",file=sys.stderr)
                    g.remove((res, SDO.targetProduct, stub))

    gpl = False
    nongpl = False
    for license in g.triples((res, SDO.license, None)):
        if str(license).startswith("http://spdx.org/licenses/GPL-") or str(license).startswith("http://spdx.org/licenses/AGPL-"):
            gpl = True
        else:
            nongpl = True
    if gpl and nongpl:
        print(f"{HEAD} license conflict, you can't use GPL alongside other licenses!",file=sys.stderr)
    print(f"{HEAD} done",file=sys.stderr)


def enrich(g: Graph, res: URIRef, args: AttribDict):
    """Do some automatic inference and enrichment of the graph"""
    IDENTIFIER = g.value(res, SDO.identifier)
    if not IDENTIFIER: IDENTIFIER = str(res)
    HEAD = f"[CODEMETA ENRICHMENT ({IDENTIFIER})]"

    if (res, SDO.targetProduct, None) not in g:
        #we have no target products, that means we have no associated interface types,
        #see if we can extract some clues from the keywords or the description
        #and add a targetproduct (with only a type)
        guess_interfacetype(g,res, args)

    if not g.value(res, SDO.programmingLanguage):
        for _,_,o in g.triples((res, SDO.runtimePlatform,None)):
            for platform in ("Python","Perl","Ruby","Julia","PHP"): #java is not added because the JVM does not necessarily mean things are written in java, NodeJS is not added because it can mean either Javascript or Typescript
                if str(o).lower().startswith(platform.lower()):
                    lang = platform
                    print(f"{HEAD} automatically adding programmingLanguage {lang} derived from runtimePlatform {platform}",file=sys.stderr)
                    g.add((res, SDO.programmingLanguage, Literal(lang)))
    elif not g.value(res, SDO.runtimePlatform):
        for _,_,o in g.triples((res, SDO.programmingLanguage,None)):
            #                                                      v--- space is needed to prevent mismatches with Javascript
            for lang in ("Python","Perl","Ruby","Julia","PHP","Java ","Kotlin","Groovy","Erlang","Elixir"):
                if str(o).lower() == lang.lower().strip() or str(o).lower().startswith(lang.lower()):
                    if lang in ("Kotlin","Groovy"):
                        platform = "Java"
                    elif lang in "Elixir":
                        platform = "Erlang"
                    else:
                        platform = lang
                    print(f"{HEAD} automatically adding runtimePlatform {platform} derived from programmingLanguage {lang}",file=sys.stderr)
                    g.add((res, SDO.runtimePlatform, Literal(platform)))

    if not g.value(res, SDO.contributor) and (res, SDO.author,None) in g:
        for _,_,o in iter_ordered_list(g, res, SDO.author):
            print(f"{HEAD} adding author {o} as contributor",file=sys.stderr)
            add_to_ordered_list(g, res, SDO.contributor, o)
    elif not g.value(res, SDO.author) and (res, SDO.contributor,None) in g:
        for _,_,o in iter_ordered_list(g, res, SDO.author):
            print(f"{HEAD} adding contributor {o} as author",file=sys.stderr)
            add_to_ordered_list(g, res, SDO.author, o)

    if not g.value(res, SDO.maintainer) and (res, SDO.author,None) in g:
        print(f"{HEAD} considering first author as maintainer",file=sys.stderr)
        for _,_,o in iter_ordered_list(g, res, SDO.author):
            g.add((res, SDO.maintainer, o))
            break

    maintainers = list(g.triples((res, SDO.maintainer,None)))
    if not g.value(res, SDO.producer) and maintainers:
        for maintainer in maintainers:
            if isinstance(maintainer, (URIRef,BNode)):
                if (maintainer, RDF.type, SDO.Person) in g:
                    for _,_,affiliation in g.triples((maintainer, SDO.affiliation,None)):
                        print(f"{HEAD} setting maintainer's affiliation as producer",file=sys.stderr)
                        g.add((res, SDO.producer, affiliation))
                elif (maintainer, RDF.type, SDO.Organization) in g:
                    print(f"{HEAD} maintainer is organization, add as producer",file=sys.stderr)
                    g.add((res, SDO.producer, maintainer))

    if not g.value(res, SDO.producer) and (res, SDO.author,None) in g:
        for _,_,author in iter_ordered_list(g, res, SDO.author):
            if isinstance(author, (URIRef,BNode)):
                if (author, RDF.type, SDO.Person) in g:
                    for _,_,o in g.triples((author, SDO.affiliation,None)):
                        print(f"{HEAD} adding affiliation(s) of first author as producer",file=sys.stderr)
                        g.add((res, SDO.producer, o))
                elif (author, RDF.type, SDO.Organization) in g:
                    print(f"{HEAD} author is organization, add as producer",file=sys.stderr)
                    g.add((res, SDO.producer, author))
            break


def guess_interfacetype(g: Graph, res: Union[URIRef,BNode], args: AttribDict) -> Union[URIRef,None]:
    IDENTIFIER = g.value(res, SDO.identifier)
    if not IDENTIFIER: IDENTIFIER = str(res)
    HEAD = f"[CODEMETA ENRICHMENT ({IDENTIFIER})]"

    counter = Counter() #we count clues for all kinds of types we can find and pick the highest one
    keywords = [ o.lower() for _,_,o in g.triples((res, SDO.keywords, None)) ]
    for clue, interfacetype in INTERFACE_CLUES:
        for keyword in keywords:
            if keyword.find(clue) != -1:
                counter.update({interfacetype:1})
    description = g.value(res, SDO.description)
    if description:
        description = description.lower()
        for clue, interfacetype in INTERFACE_CLUES:
            if description.find(clue) != -1:
                counter.update({interfacetype:1})
    #can we infer a type from the dependencies?
    for _,_, depres in g.triples((res,SDO.softwareRequirements,None)):
        depname = g.value(depres,SDO.name)
        if depname:
            depname = depname.lower()
            if depname in INTERFACE_CLUES_DEPS:
                counter.update({INTERFACE_CLUES_DEPS[depname]:1})

    if counter:
        interfacetype = max(counter)
        print(f"{HEAD} Guessing interface type {interfacetype} based on clues",file=sys.stderr)
        targetres = URIRef(generate_uri(baseuri=args.baseuri, prefix="stub"))
        g.set((targetres, RDF.type, interfacetype))
        name = g.value(res,SDO.name)
        if name:
            g.set((targetres, SDO.name, name))
        g.set((res, SDO.targetProduct, targetres))
        return targetres


def get_subgraph(g: Graph, reslist: Sequence[Union[URIRef,BNode]], subgraph: Union[Graph,None] = None, history: set = None ) -> Graph:
    """Add everything referenced from the specified resource to the new subgraph"""

    if subgraph is None:
        subgraph = Graph()
        bind_graph(subgraph)

    if history is None:
        history = set()

    for res in reslist:
        for pred, obj in g[res]:
            subgraph.add((res,pred,obj))
            if isinstance(obj, (URIRef, BNode)) and obj not in history:
                history.add(obj)
                get_subgraph(g, [obj], subgraph, history)
            elif isinstance(obj, Literal) and str(obj).startswith(("http","_","/")) and (URIRef(obj),None,None) in g and URIRef(obj) not in history: #incldue with things that are likely references but ended up as a Literal by mistake
                history.add(URIRef(obj))
                get_subgraph(g, [URIRef(obj)], subgraph, history)

    return subgraph


def getstream(source: str):
    """Opens an file (or use - for stdin) and returns the file descriptor"""
    if source == '-':
        return sys.stdin
    return open(source,'r',encoding='utf-8')


def remap_uri(g: Graph, from_uri, to_uri):
    """Changes URIs (in-graph)"""
    assert from_uri is not None and to_uri is not None
    if not isinstance(from_uri, URIRef):
        from_uri = URIRef(from_uri)
    if not isinstance(to_uri, URIRef):
        to_uri = URIRef(to_uri)
    for s,p,o in g.triples((from_uri,None,None)):
        g.remove((s,p,o))
        g.add((to_uri,p,o))
    for s,p,o in g.triples((None,None,from_uri)):
        g.remove((s,p,o))
        g.add((s,p,to_uri))


def compose(g: Graph, newgraph: Graph, res: URIRef, args: AttribDict):
    """Merges two graphs that cover the same resource (= metadata composition). Later properties will overwrite earlier ones. Newgraph will be merged into g at the end of this process."""
   
    IDENTIFIER = g.value(res, SDO.identifier) or newgraph.value(res, SDO.identifier)
    if not IDENTIFIER: IDENTIFIER = str(res)
    HEAD = f"[CODEMETA COMPOSITION ({IDENTIFIER})]"

    #Later properties always override earlier properties (they are overwritten and not merged!)
    for s,p,o in newgraph.triples((res,None,None)):
        if (s,p,None) in g:
            for s,p,o_old in g.triples((res,p,None)):
                if (s,p,o_old) not in newgraph:
                    if p in (CODEMETA.developmentStatus, SDO.applicationCategory):
                        #this is a bit simplistic for now...
                        if different_domain(o,o_old):
                            #old one an new ones are URIs and have different domain, we want to keep both, do NOT override
                            continue
                    print(f"{HEAD} overriding old {p} ({o_old} -> {o})",file=sys.stderr)
                    g.remove((s,p,o_old))

    #some correcting operations on the newgraph
    correct(newgraph, res, args)

    #there must be NO blank nodes anymore at this point!!! They might collide
    g += newgraph
    print(f"{HEAD} processed {len(newgraph)} new triples, total is now {len(g)}",file=sys.stderr)

def different_domain(res: URIRef, res2: URIRef) -> bool:
    if res.startswith("http") and res2.startswith("http"):
        return res.split('/')[2] != res2.split('/')[2]
    else:
        return False


def correct(g:Graph, res: URIRef, args: AttribDict):
    """Runs several automatic correction operations on the graph"""

    IDENTIFIER = g.value(res, SDO.identifier)
    if not IDENTIFIER: IDENTIFIER = str(res)
    HEAD = f"[CODEMETA CORRECTION ({IDENTIFIER})]"

    #when developmentStatus is a repostatus id, convert it to the full URI
    for _,_,status in g.triples((res, CODEMETA.developmentStatus,None)):
        if str(status).lower() in REPOSTATUS_MAP.values():
            print(f"{HEAD} automatically converting status {status} to repostatus URI",file=sys.stderr)
            g.remove((res, CODEMETA.developmentStatus, status))
            g.set((res, CODEMETA.developmentStatus, URIRef("https://www.repostatus.org/#" + str(status).lower())))

    #attempt to convert licenses to a full spdx.org URI
    for _,_,license in g.triples((res, SDO.license,None)):
        if str(license).upper() in ("UNKNOWN", "NOASSERTION","NONE"):
            g.remove((res, SDO.license, Literal(license)))
        elif license and isinstance(license, Literal) and not str(license).startswith("http"):
            g.remove((res, SDO.license,license))
            license = license_to_spdx(license)
            if str(license).startswith("http"):
                print(f"{HEAD} automatically converting license to spdx URI",file=sys.stderr)
                g.add((res, SDO.license, URIRef(str(license))))
            else:
                g.add((res, SDO.license, Literal(license)))
        elif isinstance(license, (Literal,URIRef)) and str(license).startswith("https://spdx.org"):
            #map to HTTP
            print(f"{HEAD} automatically converting spdx license URI from https:// to http:///",file=sys.stderr)
            remap_uri(g, license, URIRef(str(license).replace("https://","http://")))

def get_doi(g: Graph, res: Union[URIRef,BNode]) -> Optional[str]:
    """Get the DOI for a resource, looks in various places"""
    #DOI in URI?
    if str(res).startswith("https://doi.org/"):
        return str(res)[len("https://doi.org/"):]
    elif str(res).startswith("http://doi.org/"):
        return str(res)[len("http://doi.org/"):]
    #DOI in schema:identifier?
    for _,_,o in g.triples((res,SDO.identifier,None)):
        #as URL
        if str(o).startswith("https://doi.org/"):
            return str(o)[len("https://doi.org/"):]
        elif str(o).startswith("http://doi.org/"):
            return str(o)[len("http://doi.org/"):]
        elif str(o).lower().startswith("doi:"):
            return str(o)[len("doi:"):]
        elif isinstance(o, (URIRef,BNode)) and (o,RDF.type,SDO.PropertyValue) in g and str(g.value(o,SDO.propertyID)) in ("doi","DOI"):
            #as PropertyValue (recommended)
            doi = g.value(o,SDO.value)
            if doi:
                return str(doi)
    return None


def urijoin(*args) -> str:
    s = ""
    for arg in args:
        if s and s[-1] != "/": s += "/"
        s += arg
    return s

def generate_uri(identifier: Union[str,None] = None, baseuri: Union[str,None] = None, prefix: str= ""):
    """Generate an URI (aka IRI)"""
    if not identifier:
        identifier = "N" + "%032x" % random.getrandbits(128)
    else:
        identifier = identifier.lower()
        for pattern, replacement in IDENTIFIER_MAP:
            identifier = identifier.replace(pattern,replacement) #not the most efficient but it'll do
        identifier = identifier.strip("-")
    if prefix and prefix[-1] not in ('/','#'):
        prefix += '/'
    if not baseuri:
        baseuri = "file:///" #relative URI! (authority part file:// to be compatible with what rdflib assumes by default)
    elif baseuri[-1] not in ('/','#'):
        baseuri += '/'
    return baseuri + prefix + identifier

def query(g: Graph, sparql_query: str, restype=SDO.SoftwareSourceCode):
    results = []
    for result in g.query(sparql_query):
        try:
            if result.res and (result.res, RDF.type, restype) in g:
                label = g.value(result.res, SDO.name)
                if label:
                    results.append((result.res, label))
        except AttributeError:
            raise ValueError("Invalid query: Expected ?res in SPARQL query")
    results.sort(key=lambda x: x[1].lower())
    return results



import sys
import tomlkit
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
from typing import Union, IO
from codemeta.common import AttribDict, add_triple, CODEMETA, SOFTWARETYPES, add_authors, SDO, COMMON_SOURCEREPOS, SOFTWARETYPES, generate_uri


def parse_rust(g: Graph, res: Union[URIRef, BNode], file: IO ,  args: AttribDict):
    data = tomlkit.parse(file.read())
    if 'package' in data:
        for key, value in data['package'].items():
            if key == 'repository':
                add_triple(g, res, 'codeRepository', value, args)
            elif key == 'authors':
                if isinstance(value, str):
                    add_authors(g, res, value, single_author=False, baseuri=args.baseuri)
                elif isinstance(value, list):
                    for value in value:
                        add_authors(g, res, value, single_author=True, baseuri=args.baseuri)
            elif key == 'keywords':
                if isinstance(value, (list,tuple)):
                    for keyword in value:
                        add_triple(g, res, "keywords", keyword, args)
                else:
                    print("WARNING: keywords in Cargo.toml should be a list",file=sys.stderr)
            elif key == 'homepage':
                for sourcerepo in COMMON_SOURCEREPOS:
                    if value.startswith(sourcerepo) and 'repository' not in data['package']:
                        #catch if we're describing the source code repo instead
                        add_triple(g, res, "codeRepository", value, args)
                        break
                add_triple(g, res, "url", value, args)
            elif key == 'documentation':
                add_triple(g, res, "softwareHelp", value, args)
            elif key == 'categories':
                add_triple(g, res, "applicationCategory", value, args)
            elif key in ('name','description','readme','license','version'):
                add_triple(g, res, key, value, args)
    if 'dependencies' in data:
        for key, value in data['dependencies'].items():
            if isinstance(value, dict) and 'version' in value:
                add_dependency(g, res, key, value['version'], args)
            elif isinstance(value, str):
                add_dependency(g, res, key, value, args)
            
    g.add((res, SDO.programmingLanguage, Literal("Rust")))

#pylint: disable=W0621
def add_dependency(g: Graph, res: Union[URIRef, BNode], name: str, version: str, args: AttribDict):
    if version and version[0].isalnum():
        version_id = "-" + version
    else:
        version_id = version
    depres = URIRef(generate_uri(name+version_id.replace(' ',''),args.baseuri,"dependency")) #version number is deliberately in ID here!
    g.add((depres, RDF.type, SDO.SoftwareApplication))
    g.add((depres, SDO.identifier, Literal(name)))
    g.add((depres, SDO.name, Literal(name)))
    g.add((depres, SDO.version, Literal(version)))
    g.add((res, CODEMETA.softwareRequirements, depres))



#!/usr/bin/env python3

"""This library and command-line-tool converts software metadata from various metadata schemas to a generic form using codemeta (https://codemeta.github.io/) and schema.org . It also provides numerous utility functions for working with codemeta."""


# Maarten van Gompel
# CLST, Radboud University Nijmegen
# & KNAW Humanities Cluster
# GPL v3
import re
import sys
import argparse
import json
import os.path
import glob
import random
from collections import OrderedDict, defaultdict
from typing import Union, IO, Optional, Sequence, Tuple
import copy
import datetime
from pathlib import Path
import distutils.cmd #note: will be removed in python 3.12! TODO constraint <= 3.11 in apk/apt-get in Dockerfile
#pylint: disable=C0413

from rdflib import Graph, BNode, URIRef, Literal
from rdflib.namespace import RDF, OWL
from rdflib.plugins.shared.jsonld.context import Context
import rdflib.plugins.serializers.jsonld

from codemeta.common import init_graph, CODEMETA, AttribDict, getstream, SDO, reconcile, add_triple, generate_uri, remap_uri, query, enrich, compose, correct, bind_graph
import codemeta.crosswalk
import codemeta.parsers.python
import codemeta.parsers.debian
import codemeta.parsers.jsonld
import codemeta.parsers.nodejs
import codemeta.parsers.java
import codemeta.parsers.rust
import codemeta.parsers.web
import codemeta.parsers.gitapi
import codemeta.parsers.authors
import codemeta.validation
from codemeta.serializers.jsonld import serialize_to_jsonld
from codemeta.serializers.html import serialize_to_html
from codemeta.serializers.turtle import serialize_to_turtle


#class PostDevelopCommand(setuptools.command.develop.develop):
#    """Post development installation hook"""
#    def run(self):
#        setuptools_update(self)
#        super(PostDevelopCommand, self).run()
#
#class PostInstallCommand(setuptools.command.install.install):
#    """Post installation hook"""
#    def run(self):
#        setuptools_update(self)
#        super(PostInstallCommand, self).run()

class CodeMetaCommand(distutils.cmd.Command):
    description = "Generate a codemeta.json file or update an existing one, note that the package must be installed first for this to work!"
    user_options = [
        ('with-entrypoints','e','Generate entrypoints as well (custom codemeta extension not part of the official specification)'),
        ('with-stypes','t','Generate software types using targetProduct (custom extension not part of the official codemeta/schema.org specification yet)'),
        ('dry-run','n','Write to stdout instead of codemeta.json')
    ]
    def initialize_options(self):
        self.with_entrypoints = False
        self.dry_run = False

    def finalize_options(self):
        self.with_entrypoints = bool(self.with_entrypoints)
        self.dry_run = bool(self.dry_run)

    def run(self):
        """Updates the codemeta.json for this package during the setup process. Hook to be (indirectly) called from setuptools"""
        codemetafile = "codemeta.json"
        if self.dry_run:
            outputfile = "-"
        else:
            outputfile = codemetafile
        print("Writing codemeta metadata to " + outputfile,file=sys.stderr)
        if os.path.exists(codemetafile):
            build(input="json,python",output="json",outputfile=outputfile, inputsources=[codemetafile, self.distribution.metadata.name], with_entrypoints=self.with_entrypoints)
        else:
            build(input="python",output="json",outputfile=outputfile, inputsources=[self.distribution.metadata.name], with_entrypoints=self.with_entrypoints)


props, crosswalk = codemeta.crosswalk.readcrosswalk()

def main():
    """Main entrypoint for command-line usage"""

    parser = argparse.ArgumentParser(description="Converter for Python Distutils (PyPI) Metadata to CodeMeta (JSON-LD) converter. Also supports conversion from other metadata types such as those from Debian packages. The tool can combine metadata from multiple sources.")
    parser.add_argument('-t', '--with-stypes', dest="with_stypes", help="Convert entrypoints to targetProduct and classes reflecting software type (https://github.com/codemeta/codemeta/issues/#271), linking softwareSourceCode to softwareApplication or WebAPI. If enabled, any remote URLs passed to codemetapy will automatically be encoded via targetProduct.", action='store_true',required=False)
    parser.add_argument('--exact-python-version', dest="exactplatformversion", help="Register the exact python interpreter used to generate the metadata as the runtime platform. Will only register the major version otherwise.", action='store_true',required=False)
    parser.add_argument('--single-author', dest="single_author", help="CodemetaPy will attempt to check if there are multiple authors specified in the author field, if you want to disable this behaviour, set this flag", action='store_true',required=False)
    parser.add_argument('-b', '--baseuri',type=str,help="Base URI for resulting SoftwareSourceCode instances (make sure to add a trailing slash)", action='store',required=False)
    parser.add_argument('-B', '--baseurl',type=str,help="Base URL in HTML visualizations (make sure to add a trailing slash)", action='store',required=False)
    parser.add_argument('-o', '--outputtype', dest='output',type=str,help="Output type: json (default), turtle, html", action='store',required=False, default="json")
    parser.add_argument('-O','--outputfile',  dest='outputfile',type=str,help="Output file", action='store',required=False)
    parser.add_argument('-i','--inputtype', dest='inputtypes',type=str,help="Metadata input type: python, apt (debian packages), registry, json, yaml. May be a comma seperated list of multiple types if files are passed on the command line", action='store',required=False)
    parser.add_argument('-g','--graph', dest='graph',help="Output a knowledge graph that groups all input files together. Only JSON input files are supported.", action='store_true',required=False)
    parser.add_argument('-s','--select', type=str, help="Output only the selected resource (by URI) from the graph", action='store',required=False)
    parser.add_argument('-V','--validate', type=str, help="Validate against the provided SHACL file. Adds a review property with the condensed validation results.", action='store',required=False)
    parser.add_argument('--enrich', help="Enable automatic inference and enrichment of the metadata where possible", action='store_true',required=False)
    parser.add_argument('--addcontext', help="Add the specified jsonld (must be a URL) to the context (and to the context graph). May be specified multiple times.", action='append',required=False)
    parser.add_argument('--addcontextgraph', help="Add the specified jsonld or turtle (must be a URL) to the context graph, but NOT to the main json-ld context. May be specified multiple times.", action='append',required=False)
    parser.add_argument('--includecontext', help="Include all context vocabularies in the main graph and express it verbosely in serialisations. This makes the resoluting codemeta.json richer without the need to query certain external vocabularies, at the cost of added redundancy.", action='store_true',required=False)
    parser.add_argument('--interpreter', help="Start interactive python interpreter after loading the graph", action='store_true',required=False)
    parser.add_argument('--exitv', help="Set exit status according to validation result. Use with --validate", action='store_true',required=False)
    parser.add_argument('--textv', type=str, help="Set extra text to add to a validation report. Use with --validate", action='store',required=False)
    parser.add_argument('--intro', type=str, help="Set extra text (HTML) to add to the index page as an introduction", action='store',required=False)
    parser.add_argument('--css',type=str, help="Associate a CSS stylesheet (URL) with the HTML output, multiple stylesheets can be separated by a comma", action='store',  required=False)
    parser.add_argument('--no-cache',dest="no_cache", help="Do not cache context files, force redownload", action='store_true',  required=False)
    parser.add_argument('--no-extras',dest="no_extras", help="Do not include dependencies that are marked as 'extras', applies only to Python", action='store_true',  required=False)
    parser.add_argument('--codemetaserver', '--toolstore', dest="toolstore", help="When converting to HTML, link pages together for use with codemeta-server", action='store_true',  required=False)
    parser.add_argument('--strict', dest='strict', help="Strictly adhere to the codemeta standard and disable any extensions on top of it", action='store_true')
    parser.add_argument('--released', help="Signal that this software is released, this affects whether development status maps to either WIP or active", action='store_true')
    parser.add_argument('--trl', help="Attempt to add technology readiness level based on the vocabulary used by the CLARIAH project", action='store_true')
    parser.add_argument('--title', type=str, help="Title to add when generating HTML pages", action='store')
    parser.add_argument('--identifier-from-file', dest='identifier_from_file', help="Derive the identifier from the filename/module name passed to codemetapy, not from the metadata itself", action='store_true',required=False)
    parser.add_argument('inputsources', nargs='*', help='Input sources, the nature of the source depends on the type, often a file (or use - for standard input, /dev/null to start from scratch without external input), set -i accordingly with the types (must contain as many items as passed!)')

    for key, prop in sorted(props.items()):
        if key:
            parser.add_argument('--' + key,dest=key, type=str, help=prop['DESCRIPTION'] + " (Type: "  + prop['TYPE'] + ", Parent: " + prop['PARENT'] + ") [you can format the value string in json if needed]", action='store',required=False)

    args = parser.parse_args()
    if not args.strict:
        args.with_stypes = True
    if args.css:
        args.css = [ x.strip() for x in args.css.split(",") ]

    if args.baseuri and not args.baseurl:
        args.baseurl = args.baseuri

    if args.trl:
        if args.addcontext is None: args.addcontext = []
        if "https://w3id.org/research-technology-readiness-levels" not in args.addcontext:
            args.addcontext.append("https://w3id.org/research-technology-readiness-levels")

    valid = False
    if args.graph:
        #join multiple inputs into a larger graph
        g, res, args, contextgraph = read(**args.__dict__) #may deliver a res when args.select is set
    else:
        #normal behaviour
        g, res, args, contextgraph = build(**args.__dict__)
    if args.validate: 
        if res:
            valid, _ = codemeta.validation.validate(g, res, args, contextgraph)
        else:
            raise Exception("Validation can only be done on single resources, not when --graph is set and multiple are loaded/aggregated""")

    if args.includecontext:
        g += contextgraph
    output = serialize(g, res, args, contextgraph)
    if output:
        print(output)

    if args.interpreter:
        print("Starting interactive shell: variable 'g' holds the rdflib.Graph")
        import readline # optional, will allow Up/Down/History in the console
        import code
        variables = globals().copy()
        variables.update(locals())
        shell = code.InteractiveConsole(variables)
        shell.interact()

    if args.exitv and args.validate:
        return 0 if valid else 1




def serialize(g: Graph, res: Union[Sequence,URIRef,BNode,None], args: AttribDict, contextgraph: Union[Graph,None] = None, sparql_query: Optional[str] = None, **kwargs) -> str:
    if args.output == "json":
        if sparql_query: res = [ x[0]  for x in query(g, sparql_query) ]
        doc = serialize_to_jsonld(g, res, args)
        if args.outputfile and args.outputfile != "-":
            with open(args.outputfile,'w',encoding='utf-8') as fp:
                fp.write(json.dumps(doc, indent=4, ensure_ascii=False, sort_keys=True))
        else:
            return json.dumps(doc, indent=4, ensure_ascii=False, sort_keys=True)
    elif args.output in ("turtle","ttl"):
        if sparql_query: res = [ x[0]  for x in query(g, sparql_query) ]
        doc = serialize_to_turtle(g, res)
        if args.outputfile and args.outputfile != "-":
            with open(args.outputfile,'wb') as fp:
                fp.write(doc)
        else:
            return doc
    elif args.output == "html":
        if not isinstance(contextgraph, Graph):
            raise Exception("No contextgraph provided, required for HTML serialisation")
        doc = serialize_to_html(g, res, args, contextgraph, sparql_query,  **kwargs) #note: sparql query is applied in serialization function if needed
        if args.outputfile and args.outputfile != "-":
            with open(args.outputfile,'w',encoding='utf-8') as fp:
                fp.write(doc)
        else:
            return doc
    else:
        raise Exception("No such output type: ", args.output)

def reidentify(g: Graph, res: Union[URIRef,BNode], identifier: Optional[str], founduris: list, args: AttribDict) -> Union[URIRef,BNode]:
    """Reassign a new URI for the resource, or assign an original found one. The former will include a version component"""
    if founduris and not args.baseuri:
        #restore the exact original URI because we did not set a baseuri
        founduri = founduris[0]
        print(f"Remapping URI to found URI: {res} -> {founduri}",file=sys.stderr)
        remap_uri(g, res, founduri)
        res = URIRef(founduri)
    elif args.baseuri:
        for founduri in founduris:
            if not founduri.startswith("file://"): #non-local ones only
                #we've rewritten the URI, add the old one via owl:sameAs
                g.add((res, OWL.sameAs, URIRef(founduri)))
        if not args.identifier and not args.identifier_from_file:
            #see if we can find a better one from the data itself:
            identifier = get_identifier(g, res) or identifier
        if not identifier: 
            if g.value(res, SDO.name):
                identifier = str(g.value(res, SDO.name)).strip().lower().replace(" ","-").replace(":","-")
            else:
                identifier = "N" + "%032x" % random.getrandbits(128)
        version = g.value(res,SDO.version)
        if not version: version = "snapshot" #if we find no version, we append /snapshot to the URI as a version component, usually referring to the git master/main branch but not any specific version
        uri = args.baseuri + identifier + "/" + str(version)
        print(f"Remapping URI to (possibly) new identifier and version component: {res} -> {uri}",file=sys.stderr)
        remap_uri(g, res, uri)
        res = URIRef(uri)
    return res

def get_identifier(g: Graph, res: Union[URIRef,BNode])  -> Optional[str]:
    for _,_,o in g.triples((res, SDO.identifier,None)):
        if not str(o).startswith(("http://","https://")) and not ':' in str(o):
            return str(o).strip("/ ")

def read(**kwargs) -> Tuple[Graph, Union[URIRef,None], AttribDict, Graph]:
    """Read multiple resources together in a codemeta graph, and either output it all or output a selection"""

    args = AttribDict(kwargs)

    g, contextgraph = init_graph(args)

    if not args.inputsources:
        raise Exception("No inputsources specified")

    for source in args.inputsources:
        print(f"Adding json-ld file from {source} to graph",file=sys.stderr)
        codemeta.parsers.jsonld.parse_jsonld(g, None, getstream(source), args)

    #remap resource identifiers (URIs) when needed
    for s,_,_ in g.triples((None, RDF.type,SDO.SoftwareSourceCode)):
        if isinstance(s, (URIRef,BNode)):
            identifier = get_identifier(g,s)
            if isinstance(s, URIRef) and str(s).startswith("http"):
                founduris = [str(s)]
            else:
                founduris = []
            #ensure the proper URI is set
            s = reidentify(g, s, identifier, founduris, args)
            #run some automatic corrections on the graph for this resource
            correct(g, s, args)
            reconcile(g, s, args)

    if args.select:
        res = URIRef(args.select)
        if (res, None, None) not in g:
            raise KeyError("Selected resource does not exists")
    else:
        res = None

    return (g, res, args, contextgraph)


def build(**kwargs) -> Tuple[Graph, URIRef, AttribDict, Graph]:
    """Build a codemeta graph for a single resource, may be composed from different sources"""
    args = AttribDict(kwargs)

    inputsources = []
    if args.inputsources:
        inputfiles = args.inputsources
        inputtypes = args.inputtypes.split(",") if args.inputtypes else []
        guess = False
        if len(inputtypes) != len(inputfiles):
            print(f"Passed {len(inputfiles)} files/sources but specified {len(inputtypes)} input types! Automatically guessing types...",  file=sys.stderr)
            guess = True
            for inputsource in inputfiles[len(inputtypes):]:
                if inputsource == "/dev/null":
                    inputtypes.append("null")
                elif inputsource.lower().endswith("setup.py"):
                    inputtypes.append("python")
                elif inputsource.endswith("package.json"):
                    inputtypes.append("nodejs")
                elif inputsource.endswith("pyproject.toml"):
                    inputtypes.append("python")
                elif inputsource.endswith("pom.xml"):
                    inputtypes.append("java")
                elif inputsource.endswith("Cargo.toml"):
                    inputtypes.append("rust")
                elif inputsource.lower().endswith(".json") or inputsource.lower().endswith(".jsonld"):
                    inputtypes.append("json")
                elif inputsource.upper().endswith("CONTRIBUTORS"):
                    inputtypes.append("contributors")
                elif inputsource.upper().endswith("AUTHORS"):
                    inputtypes.append("authors")
                elif inputsource.upper().endswith("MAINTAINERS"):
                    inputtypes.append("maintainers")
                elif inputsource.lower().startswith("https") or inputsource.lower().startswith("git@"):
                    #test if this is a known git platform we can query via an API
                    repo_kind = codemeta.parsers.gitapi.get_repo_kind(inputsource)
                    if repo_kind:
                        inputtypes.append(repo_kind)
                    elif not inputsource.lower().startswith("git@"):
                        #otherwise it is just a website
                        inputtypes.append("web")
                elif inputsource.lower().startswith("http"):
                    inputtypes.append("web")
        inputsources = list(zip(inputfiles, inputtypes))
        if guess:
            while len(inputtypes) < len(inputfiles):
                inputfile = inputfiles[len(inputtypes)]
                print(f"No input type specified for {inputfile}, guessing this is an installed python package (may be wrong)",file=sys.stderr)
                inputtypes.append("python")
            inputsources = list(zip(inputfiles, inputtypes))
            print(f"Detected input types: {inputsources}",file=sys.stderr)
    else:
        #no input was specified
        if os.path.exists('setup.py'):
            print("No input files specified, but found python project (setup.py) in current dir, using that...",file=sys.stderr)
            print("Generating egg_info",file=sys.stderr)
            r = os.system("python3 setup.py egg_info >&2")
            #we ignore the return code for now because it may be non-zero but still have sueful results
            for path in Path('.').rglob('*.egg-info'):
                inputsources = [(".".join(str(path).split(".")[:-1]),"python")]
                break
            if not inputsources:
                if r != 0:
                    raise Exception("Could not generate egg_info (is python3 pointing to the right interpreter?)")
                raise Exception("Could not found egg_info results")
        elif os.path.exists('pyproject.toml'):
            print("No input files specified, but found python project (pyproject.toml) in current dir, using that...",file=sys.stderr)
            inputsources = [("pyproject.toml","python")]
        else:
            raise Exception("No input files specified (use - for stdin)")

    g, contextgraph = init_graph(args)


    if args.baseuri:
        args.baseuri = args.baseuri.strip('" ')
        if args.baseuri[-1] not in ('/','#','?'):
            args.baseuri += "/"
    else:
        print("Note: You did not specify a --baseuri so we will not provide identifiers (IRIs) for your SoftwareSourceCode resources (and others)", file=sys.stderr)

    #Generate a temporary ID to use for the SoftwareSourceCode resource
    #The ID will be overwritten with a more fitting one upon serialisation
    if args.identifier:
        identifier = args.identifier.strip('"')
    else:
        identifier = os.path.basename(inputsources[0][0]).lower()
    if identifier:
        identifier = identifier.replace(".codemeta.json","").replace("codemeta.json","")
        identifier = identifier.replace(".pom.xml","").replace("pom.xml","")
        identifier = identifier.replace(".package.json","").replace("package.json","")
    uri = generate_uri(identifier, args.baseuri)
    print(f"Initial URI automatically generated, may be overriden later: {uri}",file=sys.stderr)

    #add the root resource
    res = URIRef(uri)
    g.add((res, RDF.type, SDO.SoftwareSourceCode))


    founduris = [] #stores all fully qualified URIs we find fot the main resource

    l = len(inputsources)
    for i, (source, inputtype) in enumerate(inputsources):
        print(f"Processing source #{i+1} of {l}",file=sys.stderr)

        newgraph = Graph()
        bind_graph(newgraph)
          
        if inputtype == "null":
            print(f"Starting from scratch, using command line parameters to build",file=sys.stderr)
        elif inputtype == "python":
            print(f"Obtaining python package metadata for: {source}",file=sys.stderr)
            #source is a name of a package or path to a pyproject.toml file
            codemeta.parsers.python.parse_python(newgraph, res, source, crosswalk, args)
        elif inputtype == "debian":
            print(f"Parsing debian package from {source}",file=sys.stderr)
            with getstream(source) as f:
                aptlines = f.read().split("\n")
            codemeta.parsers.debian.parse_debian(newgraph, res, aptlines, crosswalk, args)
        elif inputtype == "nodejs":
            print(f"Parsing npm package.json from {source}",file=sys.stderr)
            with getstream(source) as f:
                codemeta.parsers.nodejs.parse_nodejs(newgraph, res, f, crosswalk, args)
        elif inputtype == "rust":
            print(f"Parsing rust Cargo.toml from {source}",file=sys.stderr)
            with getstream(source) as f:
                codemeta.parsers.rust.parse_rust(newgraph, res, f, args)
        elif inputtype == "java":
            print(f"Parsing java/maven pom.xml from {source}",file=sys.stderr)
            with getstream(source) as f:
                codemeta.parsers.java.parse_java(newgraph, res, f, crosswalk, args)
        elif inputtype == "json":
            print(f"Parsing json-ld file from {source}",file=sys.stderr)
            with getstream(source) as f:
                founduri = codemeta.parsers.jsonld.parse_jsonld(newgraph, res, f, args)
            if founduri and founduri not in founduris: founduris.append(founduri)
        elif inputtype == "web":
            print(f"Fallback: Obtaining metadata from remote URL {source}",file=sys.stderr)
            found = False
            for targetres in codemeta.parsers.web.parse_web(newgraph, res, source, args):
                if targetres and args.with_stypes:
                    found = True
                    print(f"Adding service (targetProduct) {source}",file=sys.stderr)
                    g.add((res, SDO.targetProduct, targetres))
            if not found:
                print(f"(no metadata found at remote URL)",file=sys.stderr)
        elif inputtype in ("github", "gitlab", "gitapi"):
            #e.g. transform git@gitlab.com/X in https://gitlab.com/X
            source = re.sub(r'git@(.*):', r'https://\1/', source)
            if source.endswith(".git"): source = source[:-4]
            if inputtype == "gitapi": #disambiguate
                inputtype = codemeta.parsers.gitapi.get_repo_kind(source)
            if inputtype:
                print(f"Querying GitAPI parser for {source}",file=sys.stderr)
                codemeta.parsers.gitapi.parse(newgraph, res, source, inputtype ,args)
            else:
                raise ValueError(f"Unable to disambiguate gitapi type")
        elif inputtype in ('authors', 'contributors','maintainers'):
            print(f"Extracting {inputtype} from {source}",file=sys.stderr)
            if inputtype == 'authors':
                prop = SDO.author
            elif inputtype == 'contributors':
                prop = SDO.contributor
            elif inputtype == 'maintainers':
                prop = CODEMETA.maintainer
            with getstream(source) as f:
                codemeta.parsers.authors.parse_authors(newgraph, res, f, args, property=prop )
        elif inputtype is not None:
            raise ValueError(f"Unknown input type: {inputtype}")

        compose(g, newgraph, res, args)

    #Process command-line arguments last
    for key in props:
        if hasattr(args, key):
            value = getattr(args, key)
            if value: value = value.strip('"')
            if value:
                add_triple(g, res, key, value, args, replace=True)

    #Reassign a new URI to the resource (if needed)
    res = reidentify(g,res, identifier, founduris, args)

    #Test and fix conflicts in the graph (and report them)
    reconcile(g, res, args)
    
    if args.enrich:
        #Some automatic infererence and enrichment
        enrich(g, res, args)

    return (g,res,args, contextgraph)



if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import sys
import os
import unittest
import json
from rdflib import Graph, BNode, URIRef, Literal
from rdflib.namespace import RDF, OWL
from codemeta.common import CODEMETA, SDO, AttribDict, SOFTWARETYPES, SOFTWAREIODATA, iter_ordered_list, SCHEMA_SOURCE, CODEMETA_SOURCE
from codemeta.codemeta import build, serialize

def debugout(g: Graph, s,p=None,o=None):
    print("DEBUG OUTPUT:", file=sys.stderr)
    for s,p,o in g.triples((s,p,o)):
        print(s,p,o, file=sys.stderr)

class BuildTest_Json(unittest.TestCase):
    """Build codemeta.json from existing codemeta.json (basically a parse, validation/reconciliation and reserialisation)"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["frog.codemeta.json"])

    def test001_sanity(self):
        """Testing whether a codemeta.json was read accurately, tests some basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)

    def test002_basics(self):
        """Testing some basic identifying properties"""
        self.assertIn( (self.res, SDO.identifier, Literal("frog")), self.g)
        self.assertIn( (self.res, SDO.name, Literal("Frog")), self.g)
        self.assertIn( (self.res, SDO.description, None), self.g) #doesn't test actual value
        self.assertIn( (self.res, SDO.version, Literal("0.26")), self.g)

    def test003_urlref(self):
        """Testing some common URL References"""
        self.assertEqual( str(self.g.value(self.res, SDO.codeRepository)),"https://github.com/LanguageMachines/frog")
        self.assertIn( (self.res, SDO.license, URIRef("http://spdx.org/licenses/GPL-3.0-only")), self.g)
        self.assertEqual( str(self.g.value(self.res, SDO.url)), "https://languagemachines.github.io/frog")

    def test004_codemeta_urlref(self):
        """Testing some codemeta URL References"""
        self.assertIn( (self.res, CODEMETA.developmentStatus, URIRef("https://www.repostatus.org/#active")), self.g)
        self.assertIn( (self.res, CODEMETA.issueTracker, URIRef("https://github.com/LanguageMachines/frog/issues")), self.g)
        self.assertIn( (self.res, CODEMETA.contIntegration, URIRef("https://travis-ci.org/LanguageMachines/frog")), self.g)
        self.assertIn( (self.res, CODEMETA.readme, URIRef("https://github.com/LanguageMachines/frog/blob/master/README.md")), self.g)

    def test005_os(self):
        """Testing operatingSystem property"""
        self.assertIn( (self.res, SDO.operatingSystem, Literal("Linux")), self.g)
        self.assertIn( (self.res, SDO.operatingSystem, Literal("BSD")), self.g)
        self.assertIn( (self.res, SDO.operatingSystem, Literal("macOS")), self.g)

    def test006_keywords(self):
        """Testing keywords property (not exhausively)"""
        self.assertIn( (self.res, SDO.keywords, Literal("nlp")), self.g)
        self.assertIn( (self.res, SDO.keywords, Literal("dutch")), self.g)

    def test007_datecreated(self):
        """Testing dateCreated property"""
        self.assertIn( (self.res, SDO.dateCreated,Literal("2011-03-31T12:35:01Z+0000",datatype=URIRef('http://schema.org/Date'))), self.g)

    def test008_authors(self):
        """Testing authors (sorted rdf list!) (not exhaustively)"""
        i = 0
        for i, (_,_,o) in enumerate(iter_ordered_list(self.g, self.res, SDO.author)):
            self.assertIn( (o, RDF.type, SDO.Person), self.g, "Testing if author is a schema:Person")
            self.assertIn( (o, SDO.givenName, None), self.g, "Testing if author has a givenName") #not testing actual value
            self.assertIn( (o, SDO.familyName, None), self.g, "Testing if author has a familyName") #not testing actual value
            self.assertIn( (o, SDO.email, None), self.g, "Testing if author has an email") #not testing actual value
        self.assertEqual(i+1, 3, "Testing number of authors")

        #testing one specific author
        self.assertIn( (URIRef("https://orcid.org/0000-0002-1046-0006"), SDO.givenName, Literal("Maarten")), self.g, "Testing specific author's givenName")
        self.assertIn( (URIRef("https://orcid.org/0000-0002-1046-0006"), SDO.familyName, Literal("van Gompel")), self.g, "Testing specific author's familyName")
        self.assertIn( (URIRef("https://orcid.org/0000-0002-1046-0006"), SDO.email, Literal("proycon@anaproy.nl")), self.g, "Testing specific author's email")

    def test009_producer(self):
        """Testing producer (not exhaustively)"""
        self.assertIn( (self.res,SDO.producer, URIRef("https://huc.knaw.nl")), self.g, "Testing producer")
        self.assertIn( (URIRef("https://huc.knaw.nl"), SDO.name, Literal("KNAW Humanities Cluster")), self.g, "Testing producer name")
        self.assertIn( (URIRef("https://huc.knaw.nl"), SDO.url, URIRef("https://huc.knaw.nl")), self.g, "Testing producer url")

    def test010_softwarehelp(self):
        """Testing softwareHelp (not exhaustively)"""
        self.assertIn( (self.res,SDO.softwareHelp, URIRef("https://frognlp.readthedocs.io")), self.g, "Testing softwareHelp")
        self.assertIn( (URIRef("https://frognlp.readthedocs.io"), RDF.type, SDO.WebSite), self.g, "Testing softwareHelp type")

    def test011_funder(self):
        funders = [ x[2] for x in self.g.triples((self.res, SDO.funder, None)) ]
        self.assertEqual(len(funders), 2, "Testing number of funders")
        for x in funders:
            self.assertIn( (x, RDF.type, SDO.Organization), self.g, "Testing if funder is a schema:Organization")

    def test012_proglang(self):
        """Testing programming language (not exhaustively)"""
        langs = [ x[2] for x in self.g.triples((self.res, SDO.programmingLanguage, None)) ]
        self.assertEqual(len(langs), 1, "Testing number of programming languages")
        self.assertIn( (langs[0],RDF.type, SDO.ComputerLanguage), self.g, "Testing programming language type")

    def test013_targetproduct(self):
        """Testing target product"""
        targetproducts = [ x[2] for x in self.g.triples((self.res, SDO.targetProduct, None)) ]
        self.assertEqual(len(targetproducts), 5, "Testing number of target products")
        found = set()
        for x in targetproducts:
            name = self.g.value(x, SDO.name)
            self.assertIsNotNone(name)
            if name == Literal("libfrog"):
                self.assertIn( (x,RDF.type, SOFTWARETYPES.SoftwareLibrary), self.g, "Testing target product type ")
                found.add(1)
            elif name == Literal("frog"):
                self.assertIn( (x,RDF.type, SOFTWARETYPES.CommandLineApplication), self.g, "Testing target product type")
                self.assertIn( (x,SOFTWARETYPES.executableName, Literal("frog")), self.g, "Testing executable name")
                self.assertIn( (x,SDO.description, None), self.g, "Testing description") #not testing actual value
                self.assertIn( (x,SDO.runtimePlatform, Literal("Linux")), self.g, "Testing runtimePlatform")
                self.assertIn( (x,SDO.runtimePlatform, Literal("BSD")), self.g, "Testing runtimePlatform")
                self.assertIn( (x,SDO.runtimePlatform, Literal("macOS")), self.g, "Testing runtimePlatform")
                found.add(2)
            else:
                self.assertIn( (x,RDF.type, SOFTWARETYPES.CommandLineApplication), self.g, "Testing target product type")
        self.assertIn(1, found)
        self.assertIn(2, found)

    def test014_iodata(self):
        """Testing IO data"""
        targetproducts = [ x[2] for x in self.g.triples((self.res, SDO.targetProduct, None)) ]
        self.assertEqual(len(targetproducts), 5, "Testing number of target products")
        found = False
        for x in targetproducts:
            name = self.g.value(x, SDO.name)
            if name == Literal("frog"):
                for _,_, y in self.g.triples((x, SOFTWAREIODATA.consumesData, None)):
                    self.assertIn( (y,RDF.type, SDO.TextDigitalDocument), self.g, "Testing consumesData type")
                    self.assertTrue( (y,SDO.encodingFormat, Literal("text/plain")) in self.g or (y,SDO.encodingFormat, Literal("application/folia+xml")) in self.g, "Testing encoding")
                    self.assertIn( (y, SDO.inLanguage, None), self.g, "Testing inLanguage type") #not testing actual value
                    found = True
        self.assertTrue(found)

    def test100_serialisation_json(self):
        """Test json serialisation"""
        s = serialize(self.g, self.res, AttribDict({ "output": "json" }), self.contextgraph)
        data = json.loads(s)
        self.assertIn(SCHEMA_SOURCE, data['@context'], "Testing schema.org in context")
        self.assertIn(CODEMETA_SOURCE, data['@context'], "Testing codemeta in context")
        self.assertEqual(data['name'], "Frog", "Testing schema:name")
        self.assertEqual(data['@type'], "SoftwareSourceCode", "Testing type")
        self.assertTrue('description' in data,  "Testing schema:description")
        self.assertEqual(data['url'], "https://languagemachines.github.io/frog", "Testing schema:url")
        self.assertEqual(data['codeRepository'], "https://github.com/LanguageMachines/frog", "Testing schema:codeRepository")
        self.assertIsInstance(data['author'], list, "Testing whether authors are in a list")
        self.assertTrue( all(isinstance(x, dict) and x['@type'] == "Person" for x in data['author']), "Testing whether all authors are schema:Person")
        self.assertIsInstance(data['developmentStatus'], list, "Testing whether we have two development statusses in a list")
        self.assertTrue( all(isinstance(x, dict) and x['@type'] == "SoftwareApplication" for x in data['softwareRequirements']), "Testing softwareRequirements")
        self.assertTrue( all(isinstance(x, dict) and x['@type'] in ("CommandLineApplication","SoftwareLibrary") for x in data['targetProduct']), "Testing targetProducts")
        self.assertTrue( all(isinstance(x, dict) and x['@type'] in ("ScholarlyArticle","TechArticle") and isinstance(x['author'],list) and x['isPartOf']['@type'] == "PublicationIssue" for x in data['referencePublication']), "Testing referencePublication")
        self.assertEqual(data['softwareHelp']['@id'], "https://frognlp.readthedocs.io", "Testing softwareHelp ID")
        self.assertEqual(data['softwareHelp']['url'], "https://frognlp.readthedocs.io", "Testing softwareHelp URL")


    def test100_serialisation_turtle(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "ttl" }), self.contextgraph)

    def test100_serialisation_html(self):
        """Test html serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "html" }), self.contextgraph)


class BuildTest_SetupPy(unittest.TestCase):

    def setUp(self):
        #relies on automatically guessing the type based on the directory we are inputsources
        os.chdir("fusus")
        try:
            self.g, self.res, self.args, self.contextgraph = build()
        finally:
            os.chdir("..")

    def test001_sanity(self):
        """Testing whether a the basics were converted accurately, tests some basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)

    def test002_basics(self):
        """Testing some basic identifying properties"""
        self.assertIn( (self.res, SDO.identifier, Literal("fusus")), self.g)
        self.assertIn( (self.res, SDO.name, Literal("fusus")), self.g)
        self.assertIn( (self.res, SDO.description, None), self.g) #doesn't test actual value
        self.assertIn( (self.res, SDO.version, Literal("0.0.2")), self.g)

    def test003_urlref(self):
        """Testing some common URL References"""
        self.assertEqual( str(self.g.value(self.res, SDO.codeRepository)), "https://github.com/among/fusus")
        self.assertIn( (self.res, SDO.license, URIRef("http://spdx.org/licenses/MIT")), self.g)
        self.assertEqual( str(self.g.value(self.res, SDO.url)), "https://github.com/among/fusus")

    def test004_codemeta_urlref(self):
        """Testing some codemeta URL References"""
        self.assertIn( (self.res, CODEMETA.developmentStatus, URIRef("https://www.repostatus.org/#wip")), self.g)

    def test006_keywords(self):
        """Testing keywords property (not exhaustively)"""
        self.assertIn( (self.res, SDO.keywords, Literal("arabic")), self.g)
        self.assertIn( (self.res, SDO.keywords, Literal("islam")), self.g)

    def test008_authors(self):
        """Testing authors (not exhaustively)"""
        i = 0
        for i, (_,_,o) in enumerate(iter_ordered_list(self.g, self.res, SDO.author)):
            self.assertIn( (o, RDF.type, SDO.Person), self.g, "Testing if author is a schema:Person")
            self.assertIn( (o, SDO.givenName, None), self.g, "Testing if author has a givenName") #not testing actual value
            self.assertIn( (o, SDO.familyName, None), self.g, "Testing if author has a familyName") #not testing actual value
        self.assertEqual(i+1, 2, "Testing number of authors")


    def test100_serialisation_json(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "json" }), self.contextgraph)

    def test100_serialisation_turtle(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "ttl" }), self.contextgraph)

    def test100_serialisation_html(self):
        """Test html serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "html" }), self.contextgraph)

class BuildTest_GithubAPI(unittest.TestCase):
    """Build codemeta.json from existing codemeta.json (basically a parse, validation/reconciliation and reserialisation)"""

    def setUp(self):
        #relies on automatically guessing the type
        #deliberately picked software that is end-of-life and will not change much anymore
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["https://github.com/proycon/labirinto"])

    def test001_api(self):
        """Testing github API response"""
        #this is a single combined test to save API queries
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)
        self.assertIn( (self.res, SDO.name, Literal("labirinto")), self.g)
        self.assertIn( str(self.g.value(self.res, SDO.codeRepository)), "https://github.com/proycon/labirinto")
        self.assertIn( str(self.g.value(self.res, CODEMETA.issueTracker)), "https://github.com/proycon/labirinto/issues")
        self.assertIn( (self.res, SDO.description, None), self.g) #doesn't test actual value
        self.assertIn( (self.res, SDO.keywords, Literal("codemeta")), self.g)

class BuildTest_JavaPomXML(unittest.TestCase):
    """Build codemeta.json from pom.xml"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["widoco.pom.xml"])

    def test001_sanity(self):
        """Testing whether a pom.xml was read accurately, tests some basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)

    def test002_basics(self):
        """Testing some basic identifying properties"""
        self.assertIn( (self.res, SDO.name, Literal("Widoco")), self.g)
        self.assertIn( (self.res, SDO.version, Literal("1.4.17")), self.g)
        self.assertIn( (self.res, SDO.runtimePlatform, Literal("Java 1.8")), self.g)

    def test100_serialisation_json(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "json" }), self.contextgraph)

    def test100_serialisation_turtle(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "ttl" }), self.contextgraph)

    def test100_serialisation_html(self):
        """Test html serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "html" }), self.contextgraph)

class BuildTest_NpmPackageJSON(unittest.TestCase):
    """Build codemeta.json from npm package json"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["labirinto.package.json"])

    def test001_sanity(self):
        """Testing whether a package.json was read accurately, tests some basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)

    def test002_basics(self):
        """Testing some basic identifying properties"""
        self.assertIn( (self.res, SDO.name, Literal("labirinto")), self.g)
        self.assertIn( (self.res, SDO.version, Literal("0.2.6")), self.g)
        self.assertIn( (self.res, SDO.runtimePlatform, Literal("npm >= 3.0.0")), self.g)
        self.assertIn( (self.res, SDO.runtimePlatform, Literal("node >= 6.0.0")), self.g)

    def test003_urlref(self):
        """Testing some common URL References"""
        self.assertEqual( str(self.g.value(self.res, SDO.codeRepository)) , "https://github.com/proycon/labirinto")
        self.assertIn( (self.res, SDO.license, URIRef("http://spdx.org/licenses/AGPL-3.0-or-later")), self.g)
        self.assertEqual( str(self.g.value(self.res, SDO.url)),  "https://github.com/proycon/labirinto")
        self.assertEqual( str(self.g.value(self.res, CODEMETA.issueTracker)),  "https://github.com/proycon/labirinto/issues")

    def test100_serialisation_json(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "json" }), self.contextgraph)

    def test100_serialisation_turtle(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "ttl" }), self.contextgraph)

    def test100_serialisation_html(self):
        """Test html serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "html" }), self.contextgraph)

class BuildTest_Web_HTML(unittest.TestCase):
    """Build codemeta.json from webpage metadata"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["https://shebanq.ancient-data.org/"], with_stypes=True)

    def test001(self):
        """Testing basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)
        service = self.g.value(self.res, SDO.targetProduct)
        self.assertIsNotNone(service)
        self.assertIn( (service, RDF.type, SDO.WebApplication), self.g)
        self.assertIn( (service, SDO.url, Literal("https://shebanq.ancient-data.org/")), self.g)
        self.assertIn( (service, SDO.name, Literal("SHEBANQ")), self.g)
        self.assertIn( (service, SDO.description, None), self.g) #not testing value
        self.assertIn( (service, SDO.author, None), self.g) #not testing value
        self.assertIn( (service, SDO.keywords, Literal("Hebrew")), self.g)
        self.assertIn( (service, SDO.keywords, Literal("Bible")), self.g)

class BuildTest_Web_JSONLD(unittest.TestCase):
    """Build codemeta.json from webpage metadata (JSON-LD)"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["https://www.delpher.nl/"], with_stypes=True)

    def test001(self):
        """Testing basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)
        service = self.g.value(self.res, SDO.targetProduct)
        self.assertIsNotNone(service)
        self.assertIn( (service, RDF.type, SDO.WebSite), self.g)
        self.assertIn( (service, SDO.url, URIRef("https://www.delpher.nl/")), self.g)
        self.assertIn( (service, SDO.potentialAction, None), self.g) #not testing value

class BuildTest_Compose(unittest.TestCase):
    """Combine two inputs for the same resource"""

    def setUp(self):
        #relies on automatically guessing the types
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["labirinto.package.json", "labirinto.codemeta-harvest.json"])

    def test001_sanity(self):
        """Testing whether a package.json was read accurately, tests some basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)

    def test002_basics(self):
        """Testing some basic identifying properties"""
        self.assertIn( (self.res, SDO.name, Literal("labirinto")), self.g)
        self.assertIn( (self.res, SDO.version, Literal("0.2.6")), self.g)
        self.assertIn( (self.res, SDO.runtimePlatform, Literal("npm >= 3.0.0")), self.g)
        self.assertIn( (self.res, SDO.runtimePlatform, Literal("node >= 6.0.0")), self.g)

    def test003_urlref(self):
        """Testing some common URL References"""
        self.assertEqual( str(self.g.value(self.res, SDO.codeRepository)) , "https://github.com/proycon/labirinto")
        self.assertIn( (self.res, SDO.license, URIRef("http://spdx.org/licenses/AGPL-3.0-or-later")), self.g)
        self.assertEqual( str(self.g.value(self.res, SDO.url)),  "https://github.com/proycon/labirinto")
        self.assertEqual( str(self.g.value(self.res, CODEMETA.issueTracker)),  "https://github.com/proycon/labirinto/issues")

    def test004_combined(self):
        """Testing properties that come from the second resource"""
        self.assertIn( (self.res, CODEMETA.developmentStatus, URIRef("https://www.repostatus.org/#unsupported")), self.g)
        self.assertIn( (self.res, CODEMETA.issueTracker, URIRef("https://github.com/proycon/labirinto/issues")), self.g)
        producer = self.g.value(self.res, SDO.producer)
        self.assertIsNotNone(producer)
        self.assertIn( (producer, RDF.type, SDO.Organization), self.g)
        self.assertIn( (producer, SDO.name, Literal("Centre for Language and Speech Technology")), self.g)
        parent = self.g.value(producer, SDO.parentOrganization)
        self.assertIsNotNone(parent)
        self.assertIn( (parent, RDF.type, SDO.Organization), self.g)
        self.assertIn( (parent, SDO.name, Literal("Centre for Language Studies")), self.g)
        grandparent = self.g.value(parent, SDO.parentOrganization)
        self.assertIsNotNone(grandparent)
        self.assertIn( (grandparent, RDF.type, SDO.Organization), self.g)
        self.assertIn( (grandparent, SDO.name, Literal("Radboud University")), self.g)

class BuildTest_Compose2(unittest.TestCase):
    """Combine multiple resources"""

    def setUp(self):
        #relies on automatically guessing the types
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["withoutid.codemeta.json", "withid.codemeta.json"])

    def test001_compose_repostatus(self):
        """Testing whether second repostatus overwrites the first one"""
        self.assertIn( (self.res, CODEMETA.developmentStatus, URIRef("https://www.repostatus.org/#active")), self.g)
        self.assertNotIn( (self.res, CODEMETA.developmentStatus, URIRef("https://www.repostatus.org/#inactive")), self.g)

    def test002_compose_orderedlist(self):
        """Testing whether second ordered list overwrites the first one"""
        self.assertEqual( len(list(self.g.triples((self.res, SDO.author, None)))), 1)
        for i, (_,_,o) in enumerate(iter_ordered_list(self.g, self.res, SDO.author)):
            self.assertIn( (o, RDF.type, SDO.Person), self.g, "Testing if author is a schema:Person")
            self.assertTrue( (o, SDO.name, Literal("John Doe")) or (o, SDO.name, Literal("Jane Doe")), "Testing if author should be in list")
        self.assertEqual(i+1, 2, "Testing number of authors")


class BuildTest_Enrich(unittest.TestCase):
    """Build codemeta.json from existing codemeta.json (basically a parse, validation/reconciliation and reserialisation)"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["frog.codemeta.json"], enrich=True)

    def test001_sanity(self):
        """Testing whether a codemeta.json was read accurately, tests some basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)

class BuildTest_RetainId(unittest.TestCase):
    """Build codemeta.json from existing codemeta.json with absolute identifier (no baseuri set)"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["withid.codemeta.json"])

    def test001_maintain_id(self):
        """Testing whether ID is retained properly"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertEqual( URIRef("http://example.org/test"), self.res, "Testing whether resource is as expected")
        self.assertIn( (URIRef("http://example.org/test"), None,None), self.g)

class BuildTest_NewId(unittest.TestCase):
    """Build codemeta.json from existing codemeta.json with identifier"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["withid.codemeta.json"],baseuri="https://tools.clariah.nl/")

    def test001_new_id(self):
        """Testing whether ID is retained properly"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertEqual( URIRef("https://tools.clariah.nl/test/0.1"), self.res, "Testing whether resource is as expected")
        self.assertIn( (URIRef("https://tools.clariah.nl/test/0.1"), None,None), self.g)
        self.assertIn( (self.res,OWL.sameAs,URIRef("http://example.org/test")), self.g, "Test if old URI is referenced via owl:sameAs")

class BuildTest_NewId2(unittest.TestCase):
    """Build codemeta.json from existing codemeta.json, but do not use found identifier, force identifier from file"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["withid.codemeta.json"],baseuri="https://tools.clariah.nl/", identifier_from_file=True)

    def test001_new_id(self):
        """Testing whether ID is retained properly"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertEqual( URIRef("https://tools.clariah.nl/withid/0.1"), self.res, "Testing whether resource is as expected")
        self.assertIn( (URIRef("https://tools.clariah.nl/withid/0.1"), None,None), self.g)
        self.assertIn( (self.res,OWL.sameAs,URIRef("http://example.org/test")), self.g, "Test if old URI is referenced via owl:sameAs")

class BuildTest_RustCargoToml(unittest.TestCase):
    """Build codemeta.json from npm package json"""

    def setUp(self):
        #relies on automatically guessing the type
        self.g, self.res, self.args, self.contextgraph = build(inputsources=["analiticcl.Cargo.toml"])

    def test001_sanity(self):
        """Testing whether a package.json was read accurately, tests some basic properties"""
        self.assertIsInstance( self.g, Graph )
        self.assertIsInstance( self.res, URIRef)
        self.assertIn( (self.res, RDF.type, SDO.SoftwareSourceCode), self.g)

    def test002_basics(self):
        """Testing some basic identifying properties"""
        self.assertIn( (self.res, SDO.name, Literal("analiticcl")), self.g)
        self.assertIn( (self.res, SDO.version, Literal("0.4.5")), self.g)
        self.assertIn( (self.res, SDO.description, None), self.g) #doesn't test actual value

    def test003_urlref(self):
        """Testing some common URL References"""
        self.assertEqual( str(self.g.value(self.res, SDO.codeRepository)) , "https://github.com/proycon/analiticcl")
        self.assertIn( (self.res, SDO.license, URIRef("http://spdx.org/licenses/GPL-3.0-or-later")), self.g)
        self.assertEqual( str(self.g.value(self.res, SDO.url)),  "https://github.com/proycon/analiticcl")
        self.assertEqual( str(self.g.value(self.res, SDO.softwareHelp)),  "https://docs.rs/analiticcl")

    def test100_serialisation_json(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "json" }), self.contextgraph)

    def test100_serialisation_turtle(self):
        """Test json serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "ttl" }), self.contextgraph)

    def test100_serialisation_html(self):
        """Test html serialisation"""
        serialize(self.g, self.res, AttribDict({ "output": "html" }), self.contextgraph)


if __name__ == '__main__':
    unittest.main()




import sys
import os.path
from datetime import datetime
from collections import OrderedDict
import codemeta.parsers.gitapi
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF, SKOS, RDFS
from typing import Union, IO, Optional, Sequence
from itertools import chain
if sys.version_info.minor < 8:
    from importlib_metadata import version as get_version #backported
else:
    from importlib.metadata import version as get_version
from codemeta.common import AttribDict, REPOSTATUS, license_to_spdx, SDO, CODEMETA, SOFTWARETYPES, SOFTWAREIODATA, TRL, CODEMETAPY, SCHEMA_SOURCE, CODEMETA_SOURCE, SCHEMA_LOCAL_SOURCE, SCHEMA_SOURCE, CODEMETA_LOCAL_SOURCE, CODEMETA_SOURCE, STYPE_SOURCE, STYPE_LOCAL_SOURCE, init_context, SINGULAR_PROPERTIES, ORDEREDLIST_PROPERTIES,  get_subgraph, get_last_component, query, iter_ordered_list, get_doi
from codemeta import __path__ as rootpath
import codemeta.parsers.gitapi
from jinja2 import Environment, FileSystemLoader

def get_triples(g: Graph, res: Union[URIRef,BNode,None], prop, labelprop=(SDO.name,SDO.legalName, RDFS.label, SKOS.prefLabel), abcsort=False, contextgraph: Optional[Graph] = None, max=0):
    """Get all triples for a particular resource and properties, also returns labels which are looked for in the contextgraph when needed, and handles sorting"""
    results = []
    havepos = False
    if not isinstance(labelprop, (tuple, list)):
        labelprop = (labelprop,)
    if res is not None and prop in ORDEREDLIST_PROPERTIES:
        triples = iter_ordered_list(g, res, prop)
    else:
        triples = g.triples((res, prop, None))
    for i, (_,_, res2) in enumerate(triples):
        if isinstance(res2, Literal) and str(res2).startswith(("http","_","/")) and (URIRef(res2),None,None) in g:
            #if a literals referers to an existing URI in the graph, treat it as a URIRef instead
            res2 = URIRef(str(res2))
        if isinstance(res2, Literal):
            results.append( (res2, res2, 9999, 9999) )
        else:
            if prop in ORDEREDLIST_PROPERTIES:
                #already returned in order
                pos = i
            else:
                #follow schema:position if available
                pos = g.value(res2, SDO.position)
                if pos is not None:
                    havepos = True
                elif isinstance(pos, int):
                    pass
                elif isinstance(pos, str):
                    pos = int(pos) if pos.isnumeric() else 9999
            label = None
            for p in labelprop:
                for _,_,candidate in g.triples((res2,p,None)):
                    if isinstance(candidate, Literal) and candidate.language in (None,'en'): #hard-coded english for now
                        label = candidate
                        break
                if label: 
                    break
                elif contextgraph:
                    for _,_,candidate in contextgraph.triples((res2,p,None)):
                        if isinstance(candidate, Literal) and candidate.language in (None,'en'): #hard-coded english for now
                            label = candidate
                            break
                    if label: 
                        break
            if label:
                results.append((label, res2, pos, _get_sortkey2(g,res2)))
            else:
                results.append((str(res2), res2, pos, _get_sortkey2(g,res2)))
        if max and len(results) >= max:
            break
    if havepos:
        try:
            results.sort(key=lambda x: x[2])
        except TypeError: #protection against edge cases, leave unsorted then
            pass
    if abcsort:
        try:
            results.sort(key=lambda x: (x[0].lower(), x[3]))
        except TypeError: #protection against edge cases, leave unsorted then
            pass
    return [ tuple(x[:2]) for x in results ]

def get_description(contextgraph: Graph, res: Union[URIRef,BNode,None]):
    """Gets the skos:note for a specific resource"""
    for _,_,candidate in contextgraph.triples((res,SKOS.note,None)):
        if isinstance(candidate, Literal) and candidate.language in (None,'en'): #hard-coded english for now
            return candidate
    return ""


def _get_sortkey2(g: Graph, res: Union[URIRef,BNode,None]):
    #set a secondary sort-key for items with the very same name
    #ensures certain interface types are listed before others in case of a tie
    if (res, RDF.type, SDO.WebApplication):
        return  0
    elif (res, RDF.type, SDO.WebSite) in g:
        return  1
    elif (res, RDF.type, SDO.WebPage) in g:
        return  3
    elif (res, RDF.type, SDO.WebAPI) in g:
        return 4
    elif (res, RDF.type, SDO.CommandLineApplication) in g:
        return 5
    else:
        return 9999


def get_index(g: Graph, restype=SDO.SoftwareSourceCode):
    groups = OrderedDict()
    for res,_,_ in g.triples((None, RDF.type, restype)):
        found = False
        label = g.value(res, SDO.name)
        if not label:
            label = g.value(res, SDO.identifier)
            if label: 
                label = label.strip("/ \n").capitalize()
            else:
                label = "~untitled"

        for _,_, group in g.triples((res, SDO.applicationSuite,None)):
            groups.setdefault((str(group),True),[]).append((res,str(label))) #explicit group
            found = True
        
        if not found:
            #ad-hoc group (singleton)
            group = str(label)
            groups.setdefault((group,False),[]).append((res,str(label))) #ad-hoc group

    for key in groups:
        groups[key].sort(key=lambda x: x[1].lower())

    return sorted(( (k[0],k[1],v) for k,v in groups.items()) , key=lambda x: x[0].lower() )

def is_resource(res) -> bool:
    return isinstance(res, (URIRef,BNode))


def get_badge(g: Graph, res: Union[URIRef,None], key):
    source =g.value(res, SDO.codeRepository).strip("/")
    cleaned_url = source
    prefix = ""
    if source.startswith("https://"):
        repo_kind = codemeta.parsers.gitapi.get_repo_kind(source)
        git_address = cleaned_url.replace('https://','').split('/')[0]
        prefix="https://"
        git_suffix=cleaned_url.replace(prefix + git_address,'')[1:]
        if "github" == repo_kind:
            #github badges
            if key == "stars":
                yield f"https://img.shields.io/github/stars/{git_suffix}.svg?style=flat&color=5c7297", None, "Stars are an indicator of the popularity of this project on GitHub"
            elif key == "issues":
                #https://shields.io/category/issue-tracking
                yield f"https://img.shields.io/github/issues/{git_suffix}.svg?style=flat&color=5c7297", None, "The number of open issues on the issue tracker"
                yield f"https://img.shields.io/github/issues-closed/{git_suffix}.svg?style=flat&color=5c7297", None, "The number of closes issues on the issue tracker"
            elif key == "lastcommits":
                yield f"https://img.shields.io/github/last-commit/{git_suffix}.svg?style=flat&color=5c7297", None, "Last commit (main branch). Gives an indication of project development activity and rough indication of how up-to-date the latest release is."
                yield f"https://img.shields.io/github/commits-since/{git_suffix}/latest.svg?style=flat&color=5c7297&sort=semver", None, "Number of commits since the last release. Gives an indication of project development activity and rough indication of how up-to-date the latest release is."
        elif "gitlab" == repo_kind:
            # https://docs.gitlab.com/ee/api/project_badges.html
            # https://github.com/Naereen/badges
            if key == "lastcommits":
                #append all found badges at the end
                encoded_git_suffix=git_suffix.replace('/', '%2F')
                response = codemeta.parsers.gitapi.rate_limit_get(f"{prefix}{git_address}/api/v4/projects/{encoded_git_suffix}/badges", "gitlab")
                if response:
                    response = response.json()
                    for badge in response:
                       if badge['kind'] == 'project': 
                        #or rendered_image_url field?
                        image_url=badge['image_url'] 
                        name=badge['name']
                        yield f"{image_url}", f"{name}", f"{name}"  
     

def has_actionable_targetproducts(g: Graph, res: Union[URIRef,BNode]) -> bool:
    for _,_,targetres in g.triples((res, SDO.targetProduct,None)):
        if (targetres, SDO.url,None) in g:
            return True
    return False

def has_displayable_targetproducts(g: Graph, res: Union[URIRef,BNode]) -> bool:
    for _,_,targetres in g.triples((res, SDO.targetProduct,None)):
        if (targetres, SDO.url,None) in g or (targetres, SDO.name,None) in g or (targetres, SOFTWARETYPES.executableName,None)  in g:
            return True
    return False
        


def type_label(g: Graph, res: Union[URIRef,None]):
    label = g.value(res, RDF.type)
    if label:
        label = label.split("/")[-1]
        return label
    else:
        return ""


def get_interface_types(g: Graph, res: Union[URIRef,None], contextgraph: Graph, fallback= False):
    """Returns labels and definitions (2-tuple) for the interface types that this SoftwareSourceCode resource provides"""
    types =  set()
    for _,_,res3 in g.triples((res, RDF.type, None)):
        if res3 != SDO.SoftwareSourceCode:
            stype =  contextgraph.value(res3, RDFS.label)
            comment = contextgraph.value(res3, RDFS.comment) #used for definitions
            if stype:
                types.add((stype,comment))
    for _,_,res2 in g.triples((res, SDO.targetProduct, None)):
        for _,_,res3 in g.triples((res2, RDF.type, None)):
            stype =  contextgraph.value(res3, RDFS.label)
            comment = contextgraph.value(res3, RDFS.comment) #used for definitions
            if stype:
                types.add((stype,comment))

    if not types and fallback:
        types.add(("Unknown", "Sorry, we don't know what kind of interfaces this software provides. No interface types have been specified or could be automatically extracted."))
    return list(sorted(types))


def get_target_platforms(g: Graph, res: Union[URIRef,None]):
    types =  set()
    for label,_ in get_triples(g, res, SDO.runtimePlatform):
        label = label.lower().split(" ")[0]
        types.add(label.capitalize())
    for label,_ in get_triples(g, res, SDO.operatingSystem):
        label = label.lower().split(" ")[0]
        types.add(label.capitalize())
    return list(sorted(types))
        
    

def serialize_to_html( g: Graph, res: Union[Sequence,URIRef,None], args: AttribDict, contextgraph: Graph, sparql_query: Optional[str] = None,  **kwargs) -> str:
    """Serialize to HTML with RDFa"""
    assert isinstance(rootpath, (list,tuple)) and len(rootpath) > 0
    env = Environment( loader=FileSystemLoader(os.path.join(rootpath[0], 'templates')), autoescape=True, trim_blocks=True, lstrip_blocks=True)
    if res and not isinstance(res, (list,tuple)):
        if (res, RDF.type, SDO.SoftwareSourceCode) in g:
            template = "page_softwaresourcecode.html"
        elif (res, RDF.type, SDO.SoftwareApplication) in g \
             or (res, RDF.type, SDO.WebPage) in g \
             or (res, RDF.type, SDO.WebSite) in g \
             or (res, RDF.type, SDO.WebAPI) in g \
             or (res, RDF.type, SOFTWARETYPES.CommandLineApplication) in g \
             or (res, RDF.type, SOFTWARETYPES.SoftwareLibrary) in g:
            template = "page_targetproduct.html"
        elif (res, RDF.type, SDO.Person) in g \
            or (res, RDF.type, SDO.Organization):
            template = "page_person_or_org.html"
        else:
            template = "page_generic.html"
        index = []
    else:
        template = kwargs.get("indextemplate","index.html")
        if isinstance(res, (list,tuple)):
            index = [("Selected resource(s)", True, [ (x, g.value(x, SDO.name)) for x in res ])]
            res = None
        elif sparql_query:
            index = query(g, sparql_query)
            index = [("Search results",True,index)]
        else:
            index = get_index(g)
    template = env.get_template(template)
    return template.render(g=g,res=res, SDO=SDO,CODEMETA=CODEMETA, CODEMETAPY=CODEMETAPY, RDF=RDF,RDFS=RDFS,STYPE=SOFTWARETYPES, SOFTWAREIODATA=SOFTWAREIODATA, REPOSTATUS=REPOSTATUS, SKOS=SKOS, TRL=TRL, get_triples=get_triples, get_description=get_description, get_target_platforms=get_target_platforms, type_label=type_label, css=args.css, contextgraph=contextgraph, URIRef=URIRef, get_badge=get_badge, now=datetime.now().strftime("%Y-%m-%d %H:%M:%S"), index=index, get_interface_types=get_interface_types,baseuri=args.baseuri,baseurl=args.baseurl, toolstore=args.toolstore, intro=args.intro, get_last_component=get_last_component, is_resource=is_resource, int=int, range=range, str=str, Literal=Literal, get_version=get_version, chain=chain,get_doi=get_doi, has_actionable_targetproducts=has_actionable_targetproducts, has_displayable_targetproducts=has_displayable_targetproducts, **kwargs)






import sys
import json
import os.path
from typing import Union, IO, Sequence, Optional
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import SKOS
from copy import copy
from codemeta.common import AttribDict, license_to_spdx, SDO, CODEMETA_SOURCE, CODEMETA_LOCAL_SOURCE, SCHEMA_SOURCE, SCHEMA_LOCAL_SOURCE, STYPE_SOURCE, STYPE_LOCAL_SOURCE, IODATA_SOURCE, IODATA_LOCAL_SOURCE, init_context, REPOSTATUS_LOCAL_SOURCE, REPOSTATUS_SOURCE, get_subgraph, PREFER_URIREF_PROPERTIES, TMPDIR, DEVIANT_CONTEXT, ORDEREDLIST_PROPERTIES

ORDEREDLIST_PROPERTIES_NAMES = list(os.path.basename(x) for x in ORDEREDLIST_PROPERTIES)
NSPREFIXES = ('schema:','codemeta:','rdf:','rdfs:','skos:','trl:','dct:','dc:','dc11:','xsd:','stype:','softwaretypes:','iodata:','softwareiodata:','owl:','og:','sh:','nwo:','repostatus:','clariah:')
#do not embed items under theme properties
NOEMBED = ("skos:broader","skos:narrower","skos:hasTopConcept","skos:inScheme", SKOS.broader, SKOS.narrower, SKOS.hasTopConcept, SKOS.inScheme)

def remove_blank_ids(data):
    """Recursively remove all blank node IDs"""
    if isinstance(data, dict):
        if '@id' in data and data['@id'].startswith("_:"):
            del data['@id']
        if 'id' in data and data['id'].startswith("_:"):
            del data['id']
        return { k: remove_blank_ids(v) for k,v in data.items() }
    elif isinstance(data, (list, tuple)):
        return [ remove_blank_ids(v) for v in data ]
    else:
        return data

def alt_sort_key(data) -> str:
    if isinstance(data, dict):
        if 'name' in data:
            return data['name']
        if '@id' in data:
            return data['@id']
        if 'identifier' in data:
            return data['identifier']
    elif isinstance(data, str):
        return data
    return "~" #just a high alphanumeric character so it ends up after normal (ascii) stuff


def sort_by_position(data: Union[list,dict,tuple,str]) -> Union[list,dict,str]:
    """If list items have a position (schema:position) index, make sure to use it for sorting. If not, sort alphabetically of name of id"""
    if isinstance(data, (list, tuple)):
        if any( isinstance(x, dict) and 'position' in x for x in data ):
            try:
                return list(sorted( ( sort_by_position(x) for x in data) , key=lambda x: x['position'] if isinstance(x, dict) and 'position' in x else 99999999 ) )
            except TypeError: #in rare cases this might fail because of some inconsistency, return unsorted then
                return [ sort_by_position(x) for x in data ]
        else:
            try:
                return list(sorted( ( sort_by_position(x) for x in data) , key=lambda x: alt_sort_key(x) ) )
            except TypeError: #in rare cases this might fail because of some inconsistency, return unsorted then
                return [ sort_by_position(x) for x in data ]
    elif isinstance(data, dict):
        if 'rdf:first' in data:
            #ordered rdf list
            return list(rdf_list_to_normal_list(data))
        else:
            for key, value in data.items():
                data[key] = sort_by_position(value)
    return data

def rdf_list_to_normal_list(data):
    if 'rdf:first' in data:
        yield data['rdf:first']
    if 'rdf:rest' in data and data['rdf:rest']:
        assert isinstance(data['rdf:rest'], dict)
        for e in rdf_list_to_normal_list(data['rdf:rest']):
            yield e


def cleanup(data: Union[dict,list,tuple,str], baseuri: Optional[str] = None) -> Union[dict,list,str]:
    """This cleans up the serialisation:
       * It Recursively removes namespace prefixes from dictionary keys
       * It removes the IDs of all former blank nodes (stubs) (making them blank again)
       * It enforces @id and @type rather than the id/type aliases
       * It removes file:// prefixes from URIs
       * It removes the _embedding_done helper
    """
    if isinstance(data, dict):
        if 'id' in data:
            data['@id'] = data['id']
            del data['id']
        if 'type' in data:
            data['@type'] = data['type']
            del data['type']
        if '_embedding_done' in data:
            del data['_embedding_done']
        if '@id' in data:
            if (data['@id'].startswith('_') or data['@id'].startswith("file://") or (baseuri and data['@id'].startswith(baseuri + "stub/"))) and len(data) > 1:
                del data['@id']
        return { key.replace('schema:','').replace('http://schema.org/','').replace('codemeta:','').replace('stypes:','') : cleanup(value, baseuri) for key, value in data.items() }
    elif isinstance(data, (list,tuple)):
        return [ cleanup(x, baseuri) for x in data ]
    elif isinstance(data,str) and data.startswith('file://'):
        return data[7:]
    else:
        return data

def find_main(data, res: Union[URIRef,None]):
    """Find the main item in the graph"""
    if '@graph' in data:
        for item in data:
            if isinstance(item, dict) and item.get("@id") == str(res):
                return item, data
        return None, None
    else:
        return data, None

def expand_implicit_id_nodes(data, idref_properties):
    """Turn nodes like `key: uri` into `key: { "@id": uri }`"""
    if isinstance(data, dict):
        for k, v in data.items():
            if k in idref_properties and isinstance(v, str) and (v.startswith(("http","_","/")) or v.startswith(NSPREFIXES)):
                data[k] = {"@id": v }
            elif k in idref_properties and isinstance(v, list):
                data[k] = [ {"@id": e } if isinstance(e, str) and (e.startswith(("http","_","/")) or e.startswith(NSPREFIXES)) else expand_implicit_id_nodes(e, idref_properties) if isinstance(e, dict) else e for e in v ]
            elif isinstance(v, dict):
                data[k] = expand_implicit_id_nodes(data[k], idref_properties)
            elif isinstance(v, list):
                data[k] = [ expand_implicit_id_nodes(e, idref_properties) if isinstance(e,dict) else e for e in v ]
    return data

def do_object_framing(data: dict, res_id: str, history: set = set(), preserve_context: bool = True):
    """JSON-LD object framing. Rdflib's json-ld serialiser doesn't implement this so we do this ourselves"""
    itemmap = {} #mapping from ids to python dicts
    if '@graph' in data:
        gather_items(data['@graph'], itemmap)
    else:
        gather_items(data, itemmap)
    #print("DEBUG itemmap", repr(itemmap))
    if res_id not in itemmap:
        raise Exception(f"Resource {res_id} not found in tree, framing not possible")
    embed_items(itemmap[res_id], itemmap, history)
    if '@context' in data and preserve_context:
        #preserve context
        itemmap[res_id]['@context'] = data['@context']
    return itemmap[res_id]
        
def gather_items(data, itemmap: dict):
    """Gather all items from a JSON-LD tree, auxiliary function for object framing"""
    if isinstance(data, list): 
        for item in data:
            gather_items(item, itemmap)
    elif isinstance(data, dict): 
        for idkey in ('@id', 'id'):
            if idkey in data and len(data) > 1 and isinstance(idkey, str):
                item_id = data[idkey]
                if idkey in itemmap:
                    #print(f"DEBUG gathered {item_id} (duplicate)",file=sys.stderr)
                    itemmap[item_id].update(data)
                else:
                    #print(f"DEBUG gathered {item_id} (new)",file=sys.stderr)
                    itemmap[item_id] = data
        for v in data.values():
            gather_items(v, itemmap)

def embed_items(data, itemmap: dict, history: set):
    """Replace all references with items, auxiliary function for object framing. The history prevents circular references."""
    if isinstance(data, list): 
        for i, item in enumerate(data):
            #print(f"DEBUG processing list, #history: {len(history)}", file=sys.stderr)
            data[i] = embed_items(item, itemmap, copy(history)) #recursion step
    elif isinstance(data, dict): 
        for idkey in ('@id', 'id'):
            if idkey in data and data[idkey] in itemmap and data[idkey] not in history:
                #print(f"DEBUG embedded {data[idkey]} (explicit)", file=sys.stderr)
                history.add(data[idkey])
                #print(f"DEBUG (recursing over embedded content)", file=sys.stderr)
                if '_embedding_done' in itemmap[data[idkey]]:
                    data = itemmap[data[idkey]]
                else:
                    data = embed_items(itemmap[data[idkey]], itemmap, copy(history))
                    data['_embedding_done'] = True
                return data
            #elif idkey in data and data[idkey] not in itemmap:
            #    print(f"DEBUG could not embed {data[idkey]}, not in graph", file=sys.stderr)
            #elif idkey in data and data[idkey] and data[idkey] in history:
            #    print(f"DEBUG could not embed {data[idkey]}, already in history, returning a reference", file=sys.stderr)
            #    return { idkey: data[idkey] }
        for k, v in data.items():
            if k not in ('@id','id') and k not in NOEMBED:
                #print(f"DEBUG processing key {k}, #history: {len(history)}", file=sys.stderr)
                data[k] = embed_items(v, itemmap, copy(history)) #recursion step
    elif isinstance(data, str) and (data.startswith(("http","file://","/","_")) or data.startswith(NSPREFIXES)) and data in itemmap and data not in history: #this is probably a reference even though it's not explicit
        #data is an URI reference we can resolve
        history.add(data)
        #print(f"DEBUG embedded {data} (implicit), recursing over embedded content", file=sys.stderr)
        if '_embedding_done' in itemmap[data]:
            data = itemmap[data]
        else:
            data = embed_items(itemmap[data], itemmap, copy(history))
            data['_embedding_done'] = True
    return data


def hide_ordered_lists(data: Union[dict,list,tuple,str], key: Optional[str] = None) -> Union[dict,list,str]:
    """Hide explicit @list nodes on known ordered list properties, on read-in they will be assumed agained via the (manipulated) context"""
    if isinstance(data, dict):
        if '@list' in data and key in ORDEREDLIST_PROPERTIES_NAMES:
            return hide_ordered_lists(data['@list'])
        else:
            for k, v in data.items():
                data[k] = hide_ordered_lists(v, k)
    elif isinstance(data, (list,tuple)):
        return [ hide_ordered_lists(v) for v in data ]
    return data


    

def rewrite_context(context, addcontext = None) -> list:
    """Rewrite local contexts to their remote counterparts"""
    if isinstance(context, list):
        for i, value in enumerate(context):
            if value == CODEMETA_LOCAL_SOURCE:
                context[i] = CODEMETA_SOURCE
            elif value == SCHEMA_LOCAL_SOURCE:
                context[i] = SCHEMA_SOURCE
            elif value == STYPE_LOCAL_SOURCE:
                context[i] = STYPE_SOURCE
            elif value == IODATA_LOCAL_SOURCE:
                context[i] = IODATA_SOURCE
            elif value == REPOSTATUS_LOCAL_SOURCE:
                context[i] = REPOSTATUS_SOURCE
            elif addcontext:
                for remote_url in addcontext:
                    if not remote_url.startswith("http"):
                        raise Exception(f"Explicitly added context (--addcontext) must be a remote URL, got {remote_url} instead")
                    local = "file://" + os.path.join(TMPDIR, os.path.basename(remote_url))
                    if value == local:
                        context[i] = remote_url
    elif isinstance(context, str):
        context = rewrite_context([context])

    if context and context[-1] == DEVIANT_CONTEXT:
        #we strip the internal 'deviant' context so it's never explicitly outputted
        context = context[:-1]
    return context

def serialize_to_jsonld(g: Graph, res: Union[Sequence,URIRef,None], args: AttribDict) -> dict:
    """Serializes the RDF graph to JSON, taking care of 'object framing' for embedded nodes"""


    #                                              v--- the internal 'deviant' context is required for the serialisation to work, it will be stripped later in rewrite_context()
    context =[ x[0] for x in init_context(args)] + [DEVIANT_CONTEXT] 
    data = json.loads(g.serialize(format='json-ld', auto_compact=True, context=context))

    #rdflib doesn't do 'object framing' so we have to do it in this post-processing step
    #if we have a single resource, it'll be the focus object the whole frame will be built around
    if res and (not isinstance(res, (list,tuple)) or len(res) == 1):
        assert isinstance(res, URIRef)
        if args.includecontext:
            data = expand_implicit_id_nodes(data, [ str(x).split("/")[-1] for x in PREFER_URIREF_PROPERTIES])
        data = do_object_framing(data, str(res))
        # Hide explicit @list nodes, on read-in they will be assumed agained via the (manipulated) context
        data = hide_ordered_lists(data)
        assert isinstance(data, dict)

        root, parent = find_main(data, res)
        if parent and len(data['@graph']) == 1 and res:
            #No need for @graph if it contains only one item now:
            assert isinstance(root, dict)
            parent.update(root)
            del data['@graph']
            root = parent
        data = sort_by_position(data)
    else:
        #we have a graph of multiple resources, structure is mostly stand-off: we do object framing on each SoftwareSourceCode instance (this does lead to some redundancy)
        if args.includecontext:
            data = expand_implicit_id_nodes(data, [ str(x).split("/")[-1] for x in PREFER_URIREF_PROPERTIES])
        if '@graph' in data:
            new_graph = []
            for item in data['@graph']:
                if isinstance(item, dict) and item.get('@type', item.get('type',None)) == 'SoftwareSourceCode':
                    item_id = item.get('@id', item.get('id', None))
                    if item_id:
                        new_graph.append(do_object_framing(data, item_id, preserve_context=False))
            data['@graph'] = new_graph
        data = hide_ordered_lists(data)
        data = sort_by_position(data)

    assert isinstance(data, dict)
    if '@context' in data:
        #remap local context references to URLs
        data['@context'] = rewrite_context(data['@context'], args.addcontext)

    #we may have some lingering prefixes which we don't need and we want @id and @type instead of 'id' and 'type', cleanup:
    data = cleanup(data, args.baseuri)

    assert isinstance(data, dict)
    return data



import sys
import json
import os.path
from typing import Union, IO
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
import lxml
from codemeta.common import AttribDict, add_triple, CODEMETA, SOFTWARETYPES, add_authors, SDO, COMMON_SOURCEREPOS, SOFTWARETYPES, license_to_spdx, generate_uri

def parse_authors(g: Graph, res: Union[URIRef, BNode], file: IO, args: AttribDict, property=SDO.author):
    """Parses a plain-text file of people, one person (author/contributor/maintainer) per line, may additionally contain e-mail addresses <> and (urls), in that order"""
    for i, line in enumerate(file.readlines()):
        line = line.strip()
        if line and line[0] != "#":
            add_authors(g, res, line, single_author=True, baseuri=args.baseuri, property=property, skip_duplicates=True, position=i+1)
